"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""
import cv2
# import cv2.cv as cv
import numpy as np
import peakutils
import copy
import fabio
import math
from skimage.morphology import white_tophat
import matplotlib.pyplot as plt
import time
from lmfit import Model, Parameters
from lmfit.models import GaussianModel
from scipy.integrate import simps
import os
import pyFAI
from os.path import  exists
from sklearn.metrics import r2_score, mean_squared_error
import collections
from ..utils.file_manager import fullPath, createFolder, getBlankImageAndMask
from ..utils.histogram_processor import *
from ..utils.image_processor import *
import pickle
import musclex

class ScanningDiffraction:
    """
    A class for Scanning Diffraction processing - go to process() to see all processing steps
    """
    def __init__(self, filepath, filename, logger = None):
        original_image = fabio.open(fullPath(filepath, filename)).data
        self.original_image = original_image
        self.filepath = filepath
        self.filename = filename
        self.logger = logger
        self.version = musclex.__version__
        self.noBGImg = getImgAfterWhiteTopHat(self.original_image)
        self.info = self.loadCache()

    def loadCache(self):
        cache_path = fullPath(self.filepath, "cp_cache")
        cache_file = fullPath(cache_path, self.filename+'.info')
        if exists(cache_file):
            info = pickle.load(open(cache_file, "rb"))
            if info != None:
                if info['program_version'] == self.version:
                    return info
        return {}

    def cacheInfo(self):
        cache_path = fullPath(self.filepath, 'cp_cache')
        createFolder(cache_path)
        cache_file = fullPath(cache_path, self.filename + '.info')
        self.info['program_version'] = self.version
        pickle.dump(self.info, open(cache_file, "wb"))

    def log(self, msg):
        """
        print and log msg to file
        :param msg: logged msg
        :return:
        """
        print(str(msg))
        if self.logger is not None:
            self.logger.debug(msg)

    def process(self, flags = {}):
        """
        All processing steps
        """
        self.updateInfo(flags)
        self.log("----------------------------------------------------------------------")
        self.log(fullPath(self.filepath, self.filename)+" is being processed ...")
        self.findCenter()
        self.get2DIntegrations()
        self.processPartialIntegrationMethod()
        self.processCentralDiffMethod()
        self.mergeRings()

        if len(self.info['merged_peaks']) > 0:
            self.fitModel()
            self.processRings()
        else:
            self.log("WARNING : no effective rings detected.")
            self.removeInfo('fitResult')
            self.removeInfo('ring_hists')
            self.removeInfo('average_ring_model')
            self.removeInfo('ring_models')
            self.removeInfo('ring_errors')

        self.log(fullPath(self.filepath, self.filename) + " has been processed.")
        self.cacheInfo()

    def removeInfo(self, k = None):
        """
        Remove information from info dictionary by k as a key. If k is None, remove all information in the dictionary
        :param k: key of dictionary
        :return: -
        """
        if k is None:
            keys = list(self.info.keys())
            for k in keys:
                del self.info[k]
        else:
            if k in self.info.keys(): # remove from dictionary if the key exists
                del self.info[k]

    def updateInfo(self, flags):
        for flag in ['ROI', 'orientation_model']:
            if flag not in self.info or flag in flags and flags[flag] != self.info[flag]:
                self.removeInfo('ring_hists')
                break
        if '90rotation' not in self.info or self.info['90rotation'] != flags['90rotation']:
            self.removeInfo('ring_hists')
        self.info.update(flags)

    def findCenter(self):
        """
        Get center of the diffraction
        :return:
        """
        if 'center' not in self.info.keys():
            img = get8bitImage(copy.copy(self.original_image))
            self.info['center'] = getCenter(img)
            self.log("Center has been calculated. center is "+str(self.info['center']))

    def get2DIntegrations(self):
        """
        Get 2D integrations and azimuthal histrogram
        :return:
        """
        if '2dintegration' not in self.info.keys():
            blank, mask = getBlankImageAndMask(self.filepath)
            img = copy.copy(self.original_image)
            noBGImg = copy.copy(self.noBGImg)
            if blank is not None:
                img = img - blank
                noBGImg = getImgAfterWhiteTopHat(img)

            center = self.info['center']

            if img.shape == (1043, 981):
                det = "pilatus1m"  # Sensor used for diffraction_mapping_bone is pilatus1m
                # rmax = max(img.shape[0] - center[1], img.shape[1] - center[0], center[0], center[1])
                rmax = max(img.shape[0] / 2, img.shape[1] / 2)
                min_endpoint = int(round(min(img.shape[0] - center[1], img.shape[1] - center[0], center[0], center[1]) *.75))
                # img = self.getFilledImage(img)
            else:
                det = "agilent_titan"  # This detector has the same size (2048,2048)
                rmax = int(round(min(img.shape[0] - center[1], img.shape[1] - center[0], center[0], center[1]) *.75))
                min_endpoint = rmax

            corners = [(0, 0), (img.shape[1], 0), (0, img.shape[0]), (img.shape[1], img.shape[0])]
            npt_rad = int(round(max([distance(center, c) for c in corners])))
            ai = pyFAI.AzimuthalIntegrator(detector=det)
            ai.setFit2D(100, center[0], center[1])

            I2D, tth, chi = ai.integrate2d(copy.copy(self.original_image), npt_rad, 360, unit="r_mm", method="csr_ocl", mask=mask)
            I2D2, tth2, chi2 = ai.integrate2d(noBGImg, npt_rad, 360, unit="r_mm", method="csr_ocl", mask=mask)

            tth_1, I = ai.integrate1d(copy.copy(self.original_image), npt_rad, unit="r_mm", mask=mask)
            tth_2, I2 = ai.integrate1d(img, npt_rad, unit="r_mm", mask=mask)

            self.info['2dintegration'] = [I2D, tth, chi]
            self.info['tophat_2dintegration'] = [I2D2, tth2, chi2]

            hists = list(I)

            if 'fixed_hull' in self.info:
                rmin = self.info['fixed_hull'][0]
                rmax = self.info['fixed_hull'][1]
            else:
                rmin = getFirstVallay(hists)

            hists = list(I2)
            self.info['rmax'] = rmax
            self.info['min_endpoint'] = min_endpoint
            hull = self.getConvexhull(np.array(hists))
            self.info['orig_hists'] = hists
            self.info['start_point'] = rmin
            self.info['hull_hist'] = hull
            self.info['area'] = simps(hull)
            if 'ROI' not in self.info:
                self.info['ROI'] = [rmin, rmax]
            self.removeInfo('m1_rings')
            self.removeInfo('m2_rings')
            self.log('2D integration has been calculated.')

    def processPartialIntegrationMethod(self): #### Method 1 ####
        """
        Process image by 1st mether : Partial Integration method
        Get radial histograms from each degree range, find peaks from each range, then merge them
        :return:
        """
        if 'm1_rings' not in self.info.keys():
            self.log("=== Partial Integration Method is being processed...")
            # Method 1: multiple conical integrations
            ref_angle = 90
            if 'partial_angle' in self.info.keys():
                ref_angle = self.info['partial_angle']

            ranges, histograms = self.get_partial_integrations(ref_angle)
            self.info['partial_ranges'] = ranges
            self.info['m1_partial_hists'] = histograms

            hulls = []
            all_peaks = []
            partial_peaks = []

            for h in histograms:
                hull = self.getConvexhull(np.array(h))
                peaks = self.findPeaksFromHist(hull)
                all_peaks.extend(peaks)
                partial_peaks.append(peaks)
                hulls.append(hull)

            self.info['m1_partial_hulls'] = hulls
            self.info['m1_partial_peaks'] = partial_peaks

            histnp = np.array(self.info['hull_hist'])
            indexes = self.select_peaks(sorted(all_peaks), times_threshold=2, distance_threshold=15)

            all_peaks_list = sorted(indexes.keys())
            peakList1 = sorted([p for p in indexes.keys() if indexes[p] >= len(ranges)/4])
            peakList2 = sorted([p for p in indexes.keys() if indexes[p] < len(ranges)/4])

            if len(peakList2) > 0:
                sigmaList = self.getInitSigma(histnp, all_peaks_list)
                normSigma = sigmaList / np.max(sigmaList)
                normHeight = [histnp[p]/ np.max(histnp) for p in all_peaks_list]
                areas = [normHeight[i] * normSigma[i] for i in range(len(normHeight))]
                normAreas = areas/np.max(areas)

                peakList2 = sorted([peakList2[i] for i in range(len(peakList2)) if normAreas[all_peaks_list.index(peakList2[i])] > 0.2])
                peakList1.extend(peakList2)

            all_peaks_list = sorted(peakList1)

            self.log("Selected peaks = " + str(all_peaks_list))
            self.info['m1_rings'] = all_peaks_list
            self.removeInfo('merged_peaks')

    def processCentralDiffMethod(self): #### Method 2 ####
        """
        Process image by 2nd method : Central difference Method
        Get 2D histogram, find the log of central difference, then find the straight line for rings
        :return:
        """
        if 'm2_rings' not in self.info.keys():
            self.log("=== Central Difference Method is being processed...")
            I2D = self.info['tophat_2dintegration'][0]
            h = 10  # Displacement of columns for difference
            central_difference = self.get_central_difference(I2D, h)
            central_difference = central_difference.clip(0, central_difference.max()) + 0.1
            central_log = np.log(central_difference)
            self.info['central_log'] = central_log

            dict_runs = self.get_runs_from_image(central_log)

            selected_rings = self.group_runs_by_ring(dict_runs)
            selected_rings = {k: v for k, v in selected_rings.items() if v[1] - v[0] >= 5}
            self.log("Selected peaks = "+ str(sorted(selected_rings.keys())))

            self.info['m2_runs_dict'] = dict_runs
            self.info['m2_central_difference'] = central_difference
            self.info['m2_rings'] = selected_rings
            self.removeInfo('merged_rings')

    def mergeRings(self):
        """
        Merge all rings together. If multiple rings have similar location, count them as one and use the average location
        :return:
        """
        if 'merged_peaks' not in self.info.keys():
            self.log("=== Merging rings ...")
            m1_rings = copy.copy(self.info['m1_rings'])
            m2_rings = copy.copy(self.info['m2_rings'])
            all_peaks_list = m1_rings
            all_peaks_list.extend(m2_rings.keys())
            all_peaks_list = self.select_peaks(all_peaks_list, distance_threshold = 15)
            merged_peaks = sorted([p for p in all_peaks_list.keys() if self.info['start_point'] < p and p <= self.info['rmax']])
            self.log("Merged rings = "+str(merged_peaks))
            self.info['merged_peaks'] = merged_peaks
            self.removeInfo('fitResult')

    def getInitSigma(self, hist, peaks):
        """
        Get Initial Sigmas of all peaks by using half-maximum height width
        :param hist: input histogram
        :param peaks: list of peaks
        :return:
        """
        sigmaList = []

        for p in peaks:
            peak_height = hist[p]
            r = 1
            l = 1
            while p-l > 0 and p+r < len(hist) and (hist[p-l] > peak_height/2. or hist[p+r] > peak_height/2.) :
                if hist[p-l] > peak_height/2.:
                    l += 1
                if hist[p+r] > peak_height/2.:
                    r += 1
            w = r+l
            sigmaList.append(w*2./2.35482)

        return sigmaList

    def fitModel(self):
        # Fit model to azimuthal integration
        if 'fitResult' not in self.info.keys():
            self.log("=== Fitting model ...")
            self.removeInfo('ring_hists')
            self.removeInfo('average_ring_model')
            self.removeInfo('ring_models')
            self.removeInfo('ring_errors')
            merged_peaks = self.info['merged_peaks']
            hists_np = np.array(self.info['hull_hist'])
            sigmaList = self.getInitSigma(hists_np, merged_peaks)
            # widthList = [selected_rings[p][1] - selected_rings[p][0] / 2 if p in selected_rings else 3 for p in merged_peaks]
            # widthList = [w if w > 0 else 3 for w in widthList]

            # x = np.array(range(len(hists_np)))

            # Fit several times and take the minimum error result
            limit_iterations = 1
            temp_model = []
            temp_error = []
            # methods = ['leastsq', 'lbfgsb', 'cg', 'tnc', 'slsqp']
            methods = ['leastsq']

            for i in range(len(methods)):
                start = time.time()
                t_m, error = fitGMMv2(hists_np, merged_peaks, sigmaList, methods[i])
                temp_model.append(t_m)
                self.log("Fitting Results = " + str(t_m))
                self.log("Error for method "+ str(methods[i])+ " = "+ str(error))
                self.log("Time for method "+ str(methods[i])+ " = "+str(time.time() - start))
                temp_error.append(error)
                if error < 0.05:
                    break

            best_ind = np.argmin(temp_error)
            result = temp_model[best_ind]
            self.log("Selected Fitting Method = "+str(methods[best_ind]))

            # Check if peaks are guessed right
            model_peaks = [result['u' + str(i)] for i in range(1, len(merged_peaks) + 1)]
            more_info = {'model_peaks': model_peaks, 'fitResult': result, 'minimize_method': methods[best_ind]}

            self.info.update(more_info)

        if 'lambda_sdd' in self.info.keys() and 'model_peaks' in self.info.keys() and len(self.info['model_peaks']) > 0:
            self.info['peak_ds'] = [(1.0 / p) * self.info['lambda_sdd'] for p in self.info['model_peaks']]

    def processRings(self):
        """
        Process rings to get informations of rings i.e. rotation, intensity
        :return:
        """
        if 'ring_hists' not in self.info.keys():
            self.log("=== Rings information is being processed ...")
            if 'orientation_model' not in self.info:
                self.info['orientation_model'] = "GMM3"
            if self.info['orientation_model'] == "Max Intensity":
                self.maxIntensityOrientation()
            elif self.info['orientation_model'] == "GMM2":
                self.processOrientation2()
            elif self.info['orientation_model'] == "GMM3":
                self.processOrientation2(model='GMM3')
            elif self.info['orientation_model'] == "Herman Factor (Half Pi)":
                self.HermanOrientation(ir='h')
            elif self.info['orientation_model'] == "Herman Factor (Pi)":
                self.HermanOrientation()

            if 'average_ring_model' in self.info:
                if '90rotation' in self.info and self.info['90rotation']:
                    self.info['average_ring_model']['u'] += np.pi/2

    def get_partial_integrations(self, ref_angle):
        """
        Get azimuthal integration by angle range
        :param ref_angle: angle range i.e. (30,60)
        :return:
        """
        histograms = []
        # Generate the angle ranges in tuples
        ranges = [(x, x + ref_angle) for x in range(0, 360, ref_angle)]
        ranges.extend([(x, x + ref_angle) for x in range(int(ref_angle / 2), 360 + int(ref_angle / 2), ref_angle)])
        ranges = sorted(ranges, key=lambda se: se[0])

        blank, mask = getBlankImageAndMask(self.filepath)
        img = copy.copy(self.original_image)
        if blank is not None:
            img = img - blank

        if img.shape == (1043, 981):
            det = "pilatus1m"  # This detector has the size (1043, 981)
        else:
            det = "agilent_titan"

        center = self.info['center']
        corners = [(0, 0), (img.shape[1], 0), (0, img.shape[0]), (img.shape[1], img.shape[0])]
        npt_rad = int(round(max([distance(center, c) for c in corners])))
        ai = pyFAI.AzimuthalIntegrator(detector=det)
        ai.setFit2D(100, center[0], center[1])

        # Compute histograms for each range
        for a_range in ranges:
            tth, I = ai.integrate1d(img, npt_rad, unit="r_mm", azimuth_range=a_range, mask=mask)
            histograms.append(I)

        return ranges, histograms

    def get_central_difference(self, I2D, h):
        # Compute central difference
        central = I2D * 2
        right_shift = -1 * np.roll(I2D, -h, axis=1)
        left_shift = -1 * np.roll(I2D, h, axis=1)
        result_r = np.add(central, right_shift)
        return np.add(result_r, left_shift)

    def getConvexhull(self, hist):
        """
        Get backgrouns subtracted histogram by applying convex hull
        :param hist:
        :param rmax:
        :return:
        """

        if 'fixed_hull' in self.info:
            start = self.info['fixed_hull'][0]
            end = self.info['fixed_hull'][1]
            return convexHull(hist, start_p=start, end_p=end)
        else:
            shist = smooth(hist, 30)
            rmax = self.info['rmax']
            start = getFirstVallay(list(hist))
            end = len(hist)
            for i in range(len(shist)-2, 0, -1):
                if shist[i+1]-shist[i] > 0:
                    end = int(round(i*0.8))
                    break
            if end-start < 100:
                end = self.info['min_endpoint']
            return convexHull(hist, start_p=start, end_p=min(end, rmax))

    def get_runs_from_image(self, central_difference):
        """
        Get runs for 2nd method
        :param central_difference:
        :return:
        """
        runs = collections.defaultdict(list)
        lenght_threshold = 60
        value_threshold = np.median(central_difference, axis=0)
        step = 10
        for i in range(step, len(value_threshold) - step):
            value_threshold[i] = np.mean(value_threshold[i - step:i + step])
        value_threshold = [1 if v < 1 else v for v in value_threshold]
        distance_threshold = 8

        for c in range(central_difference.shape[1]):
            run = []
            last_pos = 0
            for r in range(central_difference.shape[0]):
                if central_difference[r][c] > value_threshold[c]:
                    if r - last_pos > distance_threshold:
                        if len(run) > lenght_threshold:
                            line = [run[0], run[len(run) - 1]]
                            runs[c].append(line)
                        run = []

                    last_pos = r
                    run.append((r, c))
            if len(run) > lenght_threshold:
                line = [run[0], run[len(run) - 1]]
                runs[c].append(line)
        return runs

    def group_runs_by_ring(self, runs):
        """
        Find rings from input runs
        :param runs:
        :return:
        """
        result_rings = {}
        distance_threshold = 1
        i = 0
        while i < self.original_image.shape[1]:
            if i in runs:
                distance = 0
                start = i
                end = i
                for c in range(1, self.original_image.shape[1] - i):

                    if i + c in runs:
                        end = i + c
                        distance = 0
                    else:
                        distance += 1

                    if distance > distance_threshold:
                        result_rings[(end + start) / 2] = (start, end)
                        i += c
                        break
            i += 1

        return result_rings

    def findPeaksFromHist(self, orig_hist, min_dist = 30):
        """
        Get all peaks from histogram
        :param orig_hist: input histogram
        :param min_dist: if distance between two peaks is less than min_dist
        :return:
        """
        peak_list = getPeaksFromHist(orig_hist, width_thres=10)
        peak_list = movePeaks(orig_hist, peak_list)
        peak_list2 = self.select_peaks(peak_list, times_threshold=1, distance_threshold=min_dist)
        return sorted(peak_list2)

    def select_peaks(self, peaks, times_threshold = 1, distance_threshold = 10, round_val = True):
        """
        Select peaks from list of peaks . Peaks with similar location will be merge into one as average location
        :param peaks: list of peaks
        :param times_threshold: peak which its number of appears less than times_threshold will be ignored
        :param distance_threshold: minimum distance between peaks
        :return:
        """
        aux_results = {}
        results = {}

        # Create dictionary for peaks with value equal to number of matches
        for peak in peaks:
            aux_results[peak] = aux_results.setdefault(peak, 0) + 1

        # Group by same peak using distance_threshold
        l_peaks = sorted(aux_results.keys())

        prev = 0
        for i in range(len(l_peaks)):
            aux_freq = aux_results[l_peaks[i]]
            if l_peaks[i] - prev < distance_threshold:
                freq = 0
                if prev in results:
                    freq = results[prev]
                    results.pop(prev, None)

                value = 1.*(prev*freq+l_peaks[i]*aux_freq)/(freq+aux_freq)
                results[value] = aux_freq + freq
                prev = value
            else:
                results[l_peaks[i]] = aux_freq
                prev = l_peaks[i]

        if round_val:
            return {int(round(k)): v for k, v in results.items() if v >= times_threshold}
        else:
            return {k: v for k, v in results.items() if v >= times_threshold}

    def getDspacing(self, peaks):
        """
        Find D-spacing from peaks
        :param peaks: list
        :return:
        """
        if len(peaks) == 1:
            return peaks[0]

        # Number of hypothesis
        h = 4
        min_di = 1
        errors = []
        # weights = np.arange((1./(h+1)),1.,(1./(h+1)))

        for i in range(1, h + 1):
            d = peaks[0] / i
            # w = weights[i-1]
            error = 0.
            for p in peaks:
                mult = 1. * p / d
                error += abs((p-d*round(mult)))/d
            print("distance: " + str(d) + " error: " +str(error))
            errors.append(error)
            # if error < .20:
            #     return d
            if error < errors[min_di-1]:
                min_di = i

        return 1. * peaks[0] / min_di

    def get_closer_peak(self, model, num_peaks, p, distance):
        """
        Get the peak that close to p by distance
        :param model: model containing all peaks
        :param num_peaks: number of peaks
        :param p: peak location
        :param distance: minimum distance
        :return:
        """
        for i in range(1, num_peaks + 1):
            if abs(model['u' + str(i)] - p) < distance:
                return model['u' + str(i)], model['sigmad' + str(i)]
        return -1, -1

    def get_missed_rings_by_distance(self, num_peaks, distance, limit):
        """
        Get missed ring by using d-spacing
        :param num_peaks: number of peaks
        :param distance: d-spacing
        :param limit: limit number of rings
        :return:
        """
        distance_threshold = distance / 3
        peaks = []
        sigmas = []
        type = []
        sigma_default = 3
        model_result = self.info['fitResult']

        # Limit rings to inside image
        image_lenght = len(self.info['orig_hists']) - self.info['start_point']

        if limit * distance > image_lenght:
            limit = int(image_lenght/distance)

        if limit < 10:
            limit = 0
            for limit in range(1, 10):
                if limit * distance > image_lenght:
                    break

        limit = min(10 ,limit) #### jiranun

        for i in range(1, limit):
            location = distance * i
            p_found, sigma_found = self.get_closer_peak(model_result, num_peaks, location, distance_threshold)
            if p_found == -1:
                # Include non-existing peak
                peaks.append(location)
                sigmas.append(sigma_default)
                type.append({'non-existing': -1})
            else:
                peaks.append(p_found)
                sigmas.append(sigma_found)
                type.append({'existing': 0})

        return peaks, sigmas, type

    def get_ring_angle_range(self, ring_radius, img_center, shape):
        """
        Get ring angle range
        :param ring_radius: radius of ring
        :param img_center: image center
        :param shape: image shape
        :return:
        """
        angle_1 = 0
        angle_2 = 0
        x = 1
        y = 1
        # print "Center", img_center
        ref_x = min(img_center[0], shape[0] - img_center[0])
        ref_y = min(img_center[1], shape[1] - img_center[1])
        if ref_x < img_center[0]:
            x = -1
        if ref_y < img_center[1]:
            y = -1

        if ring_radius > ref_x:
            angle_1 = np.arccos(x * ref_x / ring_radius)
        if ring_radius > ref_y:
            angle_2 = y * np.arcsin(ref_y / ring_radius)
        ring_range = abs(convertRadtoDegrees(angle_1) - convertRadtoDegrees(angle_2))
        # msg = "Ring radius = "+str(ring_radius)+"  Angles = "+str(convertRadtoDegrees(angle_1))+","+str(convertRadtoDegrees(angle_2))+\
        #       "  Range = "+str(ring_range)
        # self.log(msg)
        return ring_range

    def removeValleys2(self, deep_hist, orig_hist, name = ""):
        """
        Remove valley from ring hist (pilatus line)
        :param hists: radial histogram
        :param name: -
        :return:
        """

        end_points = []
        start_points = []

        deep_hist = copy.copy(deep_hist)
        orig_hist = copy.copy(orig_hist)
        deep_smooth_hist = smooth(deep_hist,5) + 0.0000000001 # Prevent divide by zero
        i = 1

        while i < len(orig_hist)-1:
            # Find start point of a valley by its slope
            if deep_hist[i] < 0:

                # Find start point
                start_ind = i
                # for j in range(i, end_ind, -1):
                #     if j == end_ind or \
                #             ((deep_smooth_hist[j] - deep_smooth_hist[j + 1]) / deep_smooth_hist[j] < 0.0005
                #              and deep_hist[j]>0):
                #         start_ind = max(0, j-1)
                #         break

                # Move left for 2 to start before valley
                start_ind = max(0, start_ind-2)

                # Pass valley
                while i < len(deep_hist)-1 and deep_hist[i] < 0:
                    i = i + 1

                # Approximate distance from valley to end point
                # remain = 10

                # Find End point
                # for j in range(i, i+remain):
                #     end_ind = j
                #     if j >= len(orig_hist)-1 or ((deep_smooth_hist[j] - deep_smooth_hist[j-1])/deep_smooth_hist[j] < 0.0005 and deep_hist[j] > 0):
                #         end_ind = j
                #         break

                # Move right for 1 to end after valley
                end_ind = min(i+2, len(orig_hist)-1)
                i = end_ind

                if start_ind == 0 or end_ind - start_ind < 2:
                    sub = orig_hist[end_ind]
                elif end_ind == len(orig_hist)-1:
                    sub = orig_hist[start_ind]
                else:
                    xs = np.linspace(0, end_ind - start_ind + 1, end_ind - start_ind + 1)
                    val_s = orig_hist[start_ind]
                    val_e = orig_hist[end_ind]
                    m = (val_e-val_s)/len(xs)
                    sub = np.array([m*x+val_s for x in xs])

                orig_hist[start_ind: end_ind+1] = sub
                start_points.append(start_ind)
                end_points.append(end_ind)

            i = i + 1

        # print "Gab size : ", [end_points[i]-start_points[i] for i in range(len(start_points))]
        # fig = plt.figure()
        # ax = fig.add_subplot(111)
        # ax.plot(deep_hist)
        # ax.plot(start_points, deep_hist[start_points], "ro")
        # ax.plot(end_points, deep_hist[end_points], "bo")
        # fig.show()

        return orig_hist


    def removeValleys(self, hists, name = ""):
        """
        Remove valley from ring hist (pilatus line)
        :param hists: radial histogram
        :param name: -
        :return:
        """
        hist2 = copy.copy(hists)
        smooth_hist = smooth(hist2)
        thres = np.mean(hists) * 0.8
        i = 1

        while i < len(hist2):
            if hist2[i] < thres:
                start_ind = i - 1
                while i < len(hist2) and hist2[i] < thres:
                    i = i + 1

                while i < len(hist2) and ((1.*(hist2[start_ind] - hist2[i])/hist2[start_ind] > 0.1) or smooth_hist[i] - smooth_hist[i - 1] > 20. ):
                    # print name,"---",i-1,",",i,"----",smooth_hist[i] - smooth_hist[i - 1]
                    i = i + 1

                start_ind = max(0, start_ind - 3)
                end_ind = min(i, len(hist2) - 1)

                if start_ind == 0:
                    sub = hist2[end_ind]
                elif end_ind >= len(hist2) - 1:
                    sub = hist2[start_ind]
                    i = end_ind
                else:
                    xs = np.linspace(0, end_ind - start_ind + 1, end_ind - start_ind + 1)
                    val_s = smooth_hist[start_ind]
                    val_e = smooth_hist[end_ind]
                    m = (val_e-val_s)/len(xs)
                    sub = np.array([m*x+val_s for x in xs])
                hist2[start_ind: end_ind+1] = sub
            i = i + 1

        return hist2

    def process_rings(self):
        """
        Process all rings
        :return:
        """
        peaks = self.info['model_peaks']
        result = []

        center = self.info['center']
        l = int(min(self.original_image.shape[0] - center[1], self.original_image.shape[1] - center[0], center[0], center[1]))
        I2D = self.info['2dintegration'][0]

        histogram1D = self.info['hull_hist']
        rings = []
        ring_peaks = []
        ring_hists = []

        # Check for non-existing peaks and include them
        d = self.getDspacing(peaks)
        self.log("D-spacing = " + str(d))
        self.info['distance'] = d
        limit = int(np.round(max(peaks)/d * 1.5))
        rads, sigmas, type_rings = self.get_missed_rings_by_distance(len(peaks), d, limit)

        for i in range(0, len(rads)):
            h = 1
            rad_min = int(np.around(rads[i] - h * sigmas[i]))
            rad_max = int(np.around(rads[i] + h * sigmas[i]))

            if rad_max < rad_min:
                aux = rad_max
                rad_max = rad_min
                rad_min = aux
            rad_max = min(rad_max, len(histogram1D))
            rad_min = max(rad_min, 0)

            # print "Radius:", rads[i]
            # print "Integration range:", rad_max, rad_min

            # Added concern of partial rings -> divide by number of degrees
            if rads[i] > l:
                degrees = self.get_ring_angle_range(rads[i], center, self.original_image.shape)
            else:
                degrees = 360
            # Compute the area under each peak
            rings.append(simps(histogram1D[rad_min:rad_max]) / degrees)

            # Add type of ring and angle range if not complete
            if type_rings[i] != -1:
                type_rings[i] = degrees

            # Compute the azimuthal integration for each ring from 2D integration
            hists = [np.sum(I2D[j, rad_min:rad_max], axis=0) for j in range(360)]
            hists = np.array(hists)
            hists2 = copy.copy(hists)

            # Filled pilatus valleys
            if self.original_image.shape == (1043, 981):
                hists2 = self.removeValleys(hists2, str(i))

            hists = hists - min(hists2)
            hists[hists < 0] = 0
            hists2 = hists2-min(hists2)
            result.append((np.arange(0, 2 * np.pi, 2 * np.pi / 360), np.array(hists2)))

            # Compute peak of each ring
            ring_peaks.append(max(hists2))
            ring_hists.append(hists)


        self.info['rings_area'] = rings
        self.info['rings_peaks'] = ring_peaks
        self.info['radius'] = rads
        self.info['sigmas'] = sigmas
        self.info['type_rings'] = type_rings
        self.info['ring_hists'] = ring_hists
        return result

    def find_nearest(self, array, value):
        # Find index of array which has the nearest value
        idx = (np.abs(array - value)).argmin()
        return idx

    def get_ring_model(self, hist):
        """
        Fit gaussian model to rings
        :param hist:
        :return:
        """
        # Smooth histogram to find parameters easier
        # hist = (hist[0], np.convolve(hist[1], [0.5, 1, 0.5])[1:361] / 2)
        ring_hist = smooth(hist[1], 20)
        hist = (hist[0], ring_hist)

        # indexes = peakutils.indexes(np.array(hist[1]), thres=0.8, min_dist=(len(hist[1]) / 2 - 100))
        # print "INDEXES FOUND:", indexes

        # if len(indexes) == 0:
        #     return None
        #
        # u1 = hist[0][indexes[0]]
        # if len(indexes) > 2:
        #     u2 = hist[0][indexes[1]]
        #     d = indexes.sum() / 2 - indexes[0]
        # else:
        #     u2 = u1 + np.pi
        #     d = 180
        index = np.argmax(hist[1])
        u1 = hist[0][index]
        u2 = (u1 + np.pi)%(2*np.pi)
        d = 180

        # If peak is not in first half make a round sum
        min_x = index - d
        max_x = index + d
        if min_x > 0 and max_x < 360:
            alpha1 = hist[1][range(min_x, max_x)].sum() * 2 * np.pi / len(hist[1])
        else :
            if min_x < 0:
                min_x += 360
            if max_x >= 360:
                max_x -= 360
            alpha1 = hist[1][range(0, max_x)].sum() * 2 * np.pi / len(hist[1])
            alpha1 += hist[1][range(min_x, 360)].sum() * 2 * np.pi / len(hist[1])

        sigma1 = alpha1 / hist[1][index] / np.sqrt(2 * np.pi)
        sigma1 = sigma1 if not math.isnan(sigma1) else 0.5
        # Fit model using same gaussian
        x = hist[0]

        # TODO: Use the model with constrains as for fitGMMv2
        # Call orientation_GMM2
        model = Model(orientation_GMM2, independent_vars='x')
        max_height = np.max(hist[1])

        init_u = min(u1, u2)
        params = Parameters()
        params.add("u",  init_u, min=0, max=np.pi)
        params.add("sigma", sigma1, min=0, max=np.pi*2)
        params.add("alpha", alpha1, min=0, max=alpha1*5+0.0000001)
        params.add("bg", 0, min = -1, max = max_height+1)

        result = model.fit(hist[1], x=x, params = params)

        # Compute valley point and circular shift
        # print "U: ", result.values['u'], convertRadtoDegrees(result.values['u'])
        v_value = result.values['u'] + np.pi / 2
        v_point = self.find_nearest(hist[0], v_value)
        # print "Rotation point", v_point

        hist_shifted = np.roll(hist[1], -v_point)

        # Fit model again with shifted histogram
        # initial guess
        sigma = result.values['sigma']
        alpha = result.values['alpha']
        bg = result.values['bg']
        u1 = np.pi / 2.

        params = Parameters()
        params.add("u", u1, min=0, max=np.pi)
        params.add("sigma", sigma, min=0, max=np.pi*2)
        params.add("alpha", alpha, min=0, max=alpha*5+0.0000001)
        params.add("bg", bg, min = -1, max = max_height)

        model = Model(orientation_GMM2, independent_vars='x')
        result = model.fit(hist_shifted, x=x, params=params)
        result = result.values

        # Correction over shifted peaks
        result['u'] = result['u'] + v_value - np.pi

        # u2 = (u1 + np.pi) % (2 * np.pi)
        # result['u1'] = min(u1, u2)
        # result['u2'] = max(u1, u2)
        # del result['u']

        return result


    def get_ring_model2(self, hist):
        """
        Fit gaussian model to rings
        :param hist:
        :return:
        """
        # Smooth histogram to find parameters easier
        hist[1] = smooth(hist[1], 20)

        n_hist = len(hist[1])
        index = np.argmax(hist[1])
        u = hist[0][index]
        if u < np.pi / 2:
            u += np.pi
        elif u > 3 * np.pi / 2:
            u -= np.pi

        # Fit model using same gaussian
        x = hist[0]

        # Call orientation_GMM3
        model = Model(orientation_GMM3, independent_vars='x')
        max_height = np.max(hist[1])

        model.set_param_hint('u', value=u, min=np.pi/2, max=3*np.pi/2)
        model.set_param_hint('sigma', value=0.1, min=0, max=np.pi*2)
        model.set_param_hint('alpha', value=max_height*0.1/0.3989423, min=0)
        model.set_param_hint('bg', value=0, min=-1, max=max_height+1)

        result = model.fit(data=hist[1], x=x, params=model.make_params())
        errs = abs(result.best_fit - result.data)
        weights = errs / errs.mean() + 1
        weights[weights > 3.] = 0
        result = model.fit(data=hist[1], x=x, params=result.params, weights=weights)

        return result.values

    def getRingHistograms(self):
        histograms = []
        deep_valleys_hists = []

        # Filter of rings taken into account (complete)
        rings, idxs = [], []
        for i, p in enumerate(self.info['model_peaks']):
            if self.info['ROI'][0] <= p <= self.info['ROI'][1]:
                rings.append(int(round(p)))
                idxs.append(i)
        I2D = copy.copy(self.info['2dintegration'][0])
        I2D2 = copy.copy(I2D)
        if self.original_image.shape == (1043, 981):
            # if the image is from pilatus, replace gaps with large minus value to make deep valley
            I2D2[I2D2 <= 0] = -99999

        # from os.path import join
        # fabio.tifimage.tifimage(data=I2D).write(join(os.getcwd(),"I2D.tif"))
        # fig = plt.figure()
        # ax = fig.add_subplot(111)
        # ax.imshow(I2D)
        # fig.show()

        sigmas = [int(round(s)) for s in self.getInitSigma(self.info['hull_hist'], rings)]
        self.log("Peaks to check = "+ str(rings))
        for i in range(len(rings)):
            histograms.append(np.sum(I2D[:, rings[i]-sigmas[i]:rings[i]+sigmas[i]], axis=1))
            deep_valleys_hists.append(np.sum(I2D2[:, rings[i]-sigmas[i]:rings[i]+sigmas[i]], axis=1))

        if len(histograms) < 1:
            return [], [], []

        ring_hists, revised_hists = [], []

        for i in range(len(histograms)):

            # remove valleys and remove linear background
            hist = copy.copy(histograms[i])
            if self.original_image.shape == (1043, 981):
                deep_valleys_hist = deep_valleys_hists[i]
                hist = self.removeValleys2(deep_valleys_hist, hist)
            min_val = hist.min()
            hist = hist - min_val

            # Collect real histogram for displaying
            real_hist = np.array(histograms[i])
            real_hist -= min_val
            real_hist[real_hist < 0] = 0
            ring_hists.append(real_hist)
            revised_hists.append(hist)

        return ring_hists, revised_hists, idxs

    def processOrientation2(self, model='GMM2'):
        """
        Calculate ring orientation - get ring_hists, 'ring_models', 'ring_errors', 'average_ring_model'
        :return:
        """
        ring_hists, revised_hists, idx_dict = self.getRingHistograms()
        # Obtaining gaussian model of each ring histogram, means of gaussians and error over histogram
        model_dict = {}
        errors_dict = {}

        x = np.arange(0, 2 * np.pi, 2 * np.pi / 360)

        for i, hist in zip(idx_dict, revised_hists):

            # Fit orientation model
            if model == 'GMM2':
                model_dict[i] = self.get_ring_model([x, hist])
                errors_dict[i] = 1 - r2_score(orientation_GMM2(x=x, **model_dict[i]), hist)
            elif model == 'GMM3':
                model_dict[i] = self.get_ring_model2([x, hist])
                errors_dict[i] = 1 - r2_score(orientation_GMM3(x=x, **model_dict[i]), hist)

            # orig = np.array(histograms[i])
            # fig = plt.figure()
            # ax = fig.add_subplot(111)
            # ax.plot(x, real_hist)
            # ax.plot(x, hist)
            # # ax.plot(x, orig)
            # # ax.axhline(np.median(orig), label="median",color='r')
            # # ax.axhline(np.mean(orig), label="mean",color='g')
            # # ax.axhline(max(orig), label="max", color='b')
            # # ax.axhline(min(orig), label="min", color='y')
            # # ax.axhline((min(orig)+max(orig))/2., label="mid", color='m')
            # ax.plot(x, orientation_GMM2(x=x, **model_dict[i]))
            # ax.legend()
            # fig.show()

        self.info['ring_hists'] = ring_hists
        self.info['ring_models'] = model_dict
        self.info['ring_errors'] = errors_dict

        # Find avarage ranges ( ignore outliers )
        all_u1s = sorted([model_dict[i]['u'] for i in model_dict 
            if errors_dict[i] < 1. and model_dict[i]['sigma'] < 1])
        u1_dict = self.select_peaks(all_u1s, times_threshold=1, distance_threshold=0.1, round_val=False)

        # Find an average ring model ( ignore models which produce high error )
        if len(u1_dict) > 0:
            best_angle = max(u1_dict.keys(), key=lambda a:u1_dict[a])
            average_result = {
                'u' : best_angle
            }
            sum_sigma = 0
            sum_alpha = 0
            nModels = 0
            for i in model_dict:
                if errors_dict[i] < 1. and abs(model_dict[i]['u']-best_angle) < 0.1:
                    sum_sigma += model_dict[i]['sigma']
                    sum_alpha += model_dict[i]['alpha']
                    nModels += 1
            if nModels > 0:
                average_result['sigma'] = 1. * sum_sigma / nModels
                average_result['alpha'] = 1. * sum_alpha / nModels
                average_result['bg'] = 0
                self.info['average_ring_model'] = average_result
                self.log("Average Ring Model : "+ str(average_result))
            else:
                self.log("No average model detected")


    # def processOrientation(self):
    #     """
    #     Find ring orientation
    #     :return:
    #     """
    #     histograms = []
    #
    #     # Filter of rings taken into account (complete)
    #     ring_peaks = self.info['model_peaks']
    #     center = self.info['center']
    #     completeness_threshold = int(min(self.original_image.shape[0] - center[1], self.original_image.shape[1] - center[0], center[0], center[1]))
    #     self.log("Completeness threshold = "+str(completeness_threshold))
    #     self.log("Peaks to check = "+ str(ring_peaks))
    #     taken = []
    #     for i in range(len(self.info['orientations'])):
    #         if 0 < ring_peaks[i] <= completeness_threshold:
    #             histograms.append(self.info['orientations'][i])
    #             taken.append(i)
    #     self.log("Rings to use = "+str([ring_peaks[t] for t in taken]))
    #
    #     if len(histograms) < 1:
    #         return False
    #
    #     # Obtaining gaussian model of each ring histogram, means of gaussians and error over histogram
    #     model_dict = {}
    #     means_dict = {}
    #     errors_dict = {}
    #
    #     for i in range(len(histograms)):
    #
    #         model_dict[i] = self.get_ring_model(histograms[i])
    #         means_dict[i] = (model_dict[i]['u1'], model_dict[i]['u2'])
    #         errors_dict[i] = mean_squared_error(orientation_GMM2_v1(x=histograms[i][0], **model_dict[i]), histograms[i][1])
    #
    #         # plt.clf()
    #         # plt.plot(smooth(histograms[i][1], 20), color='r')
    #         # plt.plot(orientation_GMM2_v1(x=histograms[i][0], **model_dict[i]), color='g')
    #         # plt.plot(histograms[i][1], color='b')
    #         # plt.savefig("ring_"+str(i)+".png")
    #
    #
    #     # Implement the weight mechanism to decide
    #     weights = []
    #     area = self.info['rings_area']
    #     area /= max(area)
    #
    #     # max_error = max(errors_dict.values())
    #     for i in range(len(histograms)):
    #         # weights.append(abs(means_dict[i][1] - means_dict[i][0] - np.pi) * 2 * np.pi / 360)
    #         weights.append(abs(abs(means_dict[i][1] - means_dict[i][0])/np.pi-1.))
    #         # print means_dict[i]
    #         # print "First difference:", weights[i]
    #         weights[i] += errors_dict[i]
    #         # weights[i] -= area[i]
    #         # print "Total weight:", weights[i]
    #
    #     self.log("Weights = " +str(weights))
    #
    #     # Found min and use it as result
    #     min_index = np.argmin(weights)
    #
    #     hist = smooth(histograms[min_index][1], 20)
    #     mean = np.mean(hist)
    #
    #     isUniform = 1.*abs(max(hist)-mean)/mean+ 1.*abs(mean-min(hist))/mean < 1.3
    #
    #     if not isUniform:
    #         self.info['best_ring'] = taken[min_index]
    #         self.info['model'] = model_dict[min_index]

    def HoF(self, hist, mode='f'):
        """
        Calculate Herman Orientation Factors
        """
        Ints = []
        n_pi = len(hist) // 2  # number of hist unit in pi range
        n_hpi = n_pi // 2      # number of hist unit in half pi range
        for i in range(n_pi):
            I = hist[i:(i+n_pi)].copy()
            I[:i] += np.flipud(hist[:i])
            I[i:] += np.flipud(hist[(i+n_pi):])
            Ints.append(I)
        rads = np.linspace(0, np.pi, n_pi + 1)[:-1]
        denom = np.sin(rads)
        numer = (np.cos(rads)**2) * denom

        HoFs = np.zeros(hist.shape)
        for i in range(len(hist)):
            I = Ints[i] if i < n_pi else np.flipud(Ints[i - n_pi])
            if mode == 'f':
                HoFs[i] = ((I * numer).sum() / (I * denom).sum()) if i < n_pi else HoFs[i - n_pi]
            else:
                HoFs[i] = (I[:n_hpi] * numer[:n_hpi]).sum() / (I[:n_hpi] * denom[:n_hpi]).sum()
        return (3 * HoFs - 1) / 2

    def getRadOfMaxHoF(self, HoFs, mode, ratio=0.05):
        """
        Get the radian of the maximum Herman Orientation Factor
        :param HoFs:
        :param mode:
        """
        nHoFs = len(HoFs)
        num = int(nHoFs * ratio)
        if mode == 'f':
            HoFs = HoFs[:(nHoFs // 2)]
            num //= 2
        # get the indices of the top num largest HoFs
        idxs = sorted(np.arange(len(HoFs)), key=lambda i: HoFs[i])[-num:]
        idxs = sorted(idxs)
        # group the indices
        grps = [[idxs[0]]]
        for idx in idxs[1:]:
            if grps[-1][-1] == idx - 1:
                grps[-1].append(idx)
            else:
                grps.append([idx])
        # handle the round case
        if len(grps) > 1 and grps[0][0] == 0 and grps[-1][-1] == len(HoFs) - 1:
            grps[0] += [idx - len(HoFs) for idx in grps[-1]]
        # find the groups of max number of indices
        maxn = max(len(grp) for grp in grps)
        grps = [grp for grp in grps if len(grp) == maxn]
        opt_grp = sorted(grps, key=lambda g:HoFs[g].sum())[-1]
        opt_idx = np.mean(opt_grp) % len(HoFs)
        return 2 * np.pi * opt_idx / nHoFs

    def HermanOrientation(self, ir='f', thres=0.1):
        """
        Calculate Herman factors - get ring_hists,
        :param ir: integration radian, half Pi or Pi
        :return:
        """
        ROI = self.info['ROI']
        roi_hist = np.sum(self.info['2dintegration'][0][:, ROI[0]:ROI[1]], axis=1)
        roi_hist -= roi_hist.min()
        ring_hists, revised_hists, idx_dict = self.getRingHistograms()

        model_dict, errors_dict = {}, {}
        for i, hist in zip(idx_dict, revised_hists):

            # Calculate herman factors
            HoFs = self.HoF(hist, ir)
            u = self.getRadOfMaxHoF(HoFs, ir)
            model_dict[i] = {'u': u, 'HoFs': HoFs, 'sigma': 0, 'alpha': 0, 'bg': 0}
            errors_dict[i] = 1 + (HoFs.max() - thres) / (thres - 1)

        self.info['ring_hists'] = ring_hists
        self.info['ring_models'] = model_dict
        self.info['ring_errors'] = errors_dict

        roi_result = {'sigma': 0, 'alpha': 0, 'bg': 0}
        roiHoFs = self.HoF(roi_hist, ir)
        roi_result['u'] = self.getRadOfMaxHoF(roiHoFs, ir)
        roi_result['HoFs'] = roiHoFs
        self.info['average_ring_model'] = roi_result
        self.log("Herman orientation reuslt : "+ str(roi_result['u']))

    def maxIntensityOrientation(self):
        ROI = self.info['ROI']
        roi_hist = np.sum(self.info['2dintegration'][0][:, ROI[0]:ROI[1]], axis=1)
        ring_hists, revised_hists, idx_dict = self.getRingHistograms()

        model_dict, errors_dict = {}, {}
        for i, hist in zip(idx_dict, revised_hists):

            # Calculate herman factors
            u = max(np.arange(180), key=lambda d: np.sum(hist[d:d + 1]) + np.sum(hist[d + 180:d + 181])) * np.pi / 180
            model_dict[i] = {'u': u, 'sigma': 0, 'alpha': 0, 'bg': 0}
            errors_dict[i] = 0

        self.info['ring_hists'] = ring_hists
        self.info['ring_models'] = model_dict
        self.info['ring_errors'] = errors_dict

        roi_result = {'sigma': 0, 'alpha': 0, 'bg': 0, 'hist': roi_hist}
        roi_result['u'] = max(np.arange(180), key=lambda d: np.sum(roi_hist[d:d + 1]) + \
            np.sum(roi_hist[d + 180:d + 181])) * np.pi / 180
        self.info['average_ring_model'] = roi_result
        self.log("Max intensity orientation : "+ str(roi_result['u']))

############################# Batch mode Results #########################################

def toFloat(value):
    try:
        x = float(value)
    except Exception:
        x = 0

    return x

def convertRadtoDegrees(rad):
    rad = toFloat(rad)
    return int(rad * 180 / np.pi)

# def update_int(val):
#     global intensity, int_displayi, fig, quad, ax1
#     if sbluri.val > 2:
#         b1 = int(sbluri.val)
#         int_displayi = copy.copy(intensity)
#         int_displayi = cv2.blur(int_displayi, (b1, b1))
#     else:
#         int_displayi = copy.copy(intensity)
#     fig = plt.figure("Total intensity map")
#     ax1 = plt.subplot(111)
#     plt.subplots_adjust(left=0.15, bottom=0.25)
#     quad = plt.pcolormesh(x_coor, y_coor, int_displayi)

##########################################################################################


######### Gaussian Models #############
def GM(x, u1, sigmad1, alpha1):
    return alpha1 * np.exp(-1. * (x - u1) ** 2 / (2 * sigmad1 ** 2)) * (1. / (np.sqrt(2 * np.pi) * sigmad1))

def GMM_any(x, params):
    n = int(len(params.keys())/3)
    result = GM(x, params['u1'], params['sigmad1'], params['alpha1'])
    for i in range(2, n+1):
        result += GM(x, params['u'+str(i)], params['sigmad'+str(i)], params['alpha'+str(i)])
    return result


def orientation_GMM2(x, u, sigma, alpha, bg):
    mod = GaussianModel()
    return mod.eval(x=x, amplitude=alpha, center=u, sigma=sigma) + mod.eval(x=x, amplitude=alpha, center=u+np.pi, sigma=sigma) + bg

def orientation_GMM2_v1(x, u1, u2, sigma, alpha):
    mod = GaussianModel()
    return mod.eval(x=x, amplitude=alpha, center=u1, sigma=sigma) + mod.eval(x=x, amplitude=alpha, center=u2, sigma=sigma)


def orientation_GMM3(x, u, sigma, alpha, bg):
    mod = GaussianModel()
    return mod.eval(x=x, amplitude=alpha, center=u, sigma=sigma) + \
        mod.eval(x=x, amplitude=alpha, center=u-np.pi, sigma=sigma) + \
        mod.eval(x=x, amplitude=alpha, center=u+np.pi, sigma=sigma) + bg

def fitGMMv2(hists_np, indexes, widthList, method='leastsq'):
    parameters = []
    indexes = copy.copy(indexes)

    for index, width in zip(indexes, widthList):
        sigmad = 1. / np.sqrt(8 * np.log(2)) * width
        alpha = sigmad * hists_np[index] * np.sqrt(2 * np.pi)
        parameters.append((sigmad, alpha))

    x = np.array(range(len(hists_np)))

    pars = Parameters()
    gaussians = []
    mean_margin = 15
    for i in range(len(indexes)):
        prefix = 'g' + str(i + 1) + '_'
        gauss = GaussianModel(prefix=prefix)
        pars.update(gauss.make_params())
        pars[prefix + 'center'].set(indexes[i], min=indexes[i] - mean_margin, max=indexes[i] + mean_margin)
        pars[prefix + 'sigma'].set(parameters[i][0], min=1., max=15.)
        pars[prefix + 'amplitude'].set(parameters[i][1], min=1, max=1000 * hists_np[indexes[i]])
        gaussians.append(gauss)


    model = gaussians[0]
    for i in range(1, len(gaussians)):
        model += gaussians[i]

    out = model.fit(hists_np, pars, x=x, method=method).values
    # print "Output Params:", pars
    result = {}
    for i in range(len(indexes)):
        prefix = 'g' + str(i + 1) + '_'
        result['u' + str(i + 1)] = out[prefix + 'center']
        result['sigmad' + str(i + 1)] = out[prefix + 'sigma']
        result['alpha' + str(i + 1)] = out[prefix + 'amplitude']

    error = r2_score(GMM_any(x = x, params = result), hists_np)

    # if len(indexes) == 2:
    #     error = mean_square_error(GMM2(x=x, **result), hists_np)
    # if len(indexes) == 3:
    #     error = mean_square_error(GMM3(x=x, **result), hists_np)
    # if len(indexes) == 4:
    #     error = mean_square_error(GMM4(x=x, **result), hists_np)
    # if len(indexes) == 5:
    #     error = mean_square_error(GMM5(x=x, **result), hists_np)
    # if len(indexes) == 6:
    #     error = mean_square_error(GMM6(x=x, **result), hists_np)
    # if len(indexes) == 7:
    #     error = mean_square_error(GMM7(x=x, **result), hists_np)
    # if len(indexes) == 8:
    #     error = mean_square_error(GMM8(x=x, **result), hists_np)
    # if len(indexes) == 9:
    #     error = mean_square_error(GMM9(x=x, **result), hists_np)
    # if len(indexes) == 10:
    #     error = mean_square_error(GMM10(x=x, **result), hists_np)

    return result, error

def gaussian(x, a, mean, sigma):
    return a*np.exp((-1.*(x-mean)**2)/(2*(sigma**2)))

####################################

def mean_square_error(y_predict, y):
    """Find error"""
    loss = (y_predict - y)
    # nom = y - mean(y)
    # print loss
    mse = np.dot(loss.transpose(), loss) / y.shape[0]
    rmse = np.sqrt(mse)
    return rmse/(max(y)-min(y))


###### White top hat image for Scanning Diffraction #########
def getImgAfterWhiteTopHat(img, sigma=5):
    tmpKernel = 1. / sigma ** 2 * np.ones((sigma, sigma))
    dst = copy.copy(img)
    dst = np.array(dst, np.float32)
    for _ in range(2):
        dst = cv2.filter2D(dst, cv2.CV_32F, tmpKernel, anchor=(-1, -1))

    sigma = sigma * 6
    x = np.array(range(-int(sigma + 1), int(sigma + 1) + 1, 1))
    kernelX = gaussian(x, 1, 0, sigma)
    kernelXY = np.outer(kernelX, np.transpose(kernelX))
    # kernelXY = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (int(sigma), int(sigma)))
    tophat = white_tophat(dst, kernelXY)
    return tophat

def kernelXY(sigma=5):
    a = sigma * 6
    x = np.array(range(-int(a + 1), int(a + 1) + 1, 1))
    kernelX = 1. / (np.sqrt(2. * np.pi) * a) * np.exp(-(x - 0) ** 2. / (2. * a ** 2))
    return np.outer(kernelX, np.transpose(kernelX))

def whiteTopHat(img, sigma=5, first=True):
    tmpKernel = 1. / sigma ** 2 * np.ones((sigma, sigma))
    if first:
        dst = copy.copy(img)
        # print dst.max()
        # low_values_indices = dst < 0
        # dst[low_values_indices] = -1000
        dst = np.array(dst, np.float32)
        for _ in range(2):
            dst = cv2.filter2D(dst, cv2.CV_32F, tmpKernel, anchor=(-1, -1))
        kernelxy = kernelXY()
        tophat = white_tophat(dst, kernelxy)
        low_values_indices = tophat < 0
        if True in low_values_indices:
            print("There are negative indexes after tophat")
        else:
            print("There are NOT negative values after tophat")
        return tophat
    else:
        dst = copy.copy(img)
        for _ in range(2):
            dst = cv2.filter2D(dst, -1, tmpKernel, anchor=(-1, -1))
        kernelxy = kernelXY()
        tophat = white_tophat(dst, kernelxy)
        dst = copy.copy(tophat)
        for _ in range(2):
            dst = cv2.filter2D(dst, -1, tmpKernel, anchor=(-1, -1))
        kernelxy = kernelXY(sigma)
        tophat = white_tophat(dst, kernelxy)
        return tophat
############################################



"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""

from os.path import exists
from os import makedirs
import pandas as pd
from ..utils.file_manager import fullPath

class EQ_CVSManager:
    """
    A class taking care of writing results including csv file and failedcases file
    """
    def __init__(self, dir_path):
        """
        init with directory path
        :param dir_path:
        """
        result_path = fullPath(dir_path, "eq_results")
        if not exists(result_path):
            makedirs(result_path)
        self.filename = fullPath(result_path, 'summary.csv')
        self.colnames = ["Filename","Side","Distance From Center","Area","Sigma D","Sigma S","Sigma C","gamma","Z line","Sigma Z","Iz","gamma Z", "Model", "CenterX", "S10", "d10", "I11/I10", "Avarage I11/I10 per fiber", "Fitting error","comment"]
        self.loadFailedCases(dir_path)
        self.loadSummary()

    def loadFailedCases(self, dir):
        """
        Load failed cases file from the directory and keep t2hem in self.failedcases
        :param dir: input directory (str)
        :return: -
        """
        self.failedcasesfile = fullPath(dir, "failedcases.txt")
        self.failedcases = set()
        if exists(self.failedcasesfile):
            for line in open(self.failedcasesfile, 'r'):
                name = line.rstrip('\n')
                self.failedcases.add(name)

    def loadSummary(self):
        """
        Load summary.csv file and keep data in self.dataframe
        :return:
        """
        if not exists(self.filename):
            self.dataframe = pd.DataFrame(columns = self.colnames )
        else:
            self.dataframe = pd.read_csv(self.filename)
        # print self.dataframe

    def writeNewData(self, bioImg):
        """
        Add new data to dataframe, then re-write summary.csv and failed cases file
        :param bioImg: EquatorImage object with results in its info dict
        :return: -
        """
        file_name = bioImg.filename
        info = bioImg.info
        self.removeData(file_name)
        new_datas = []

        # If image is rejected
        if "reject" in info and info["reject"]:
            data = {}
            for k in self.dataframe.columns:
                data[k] = '-'
            data['Filename'] = file_name
            data['comment'] = "REJECTED"
            new_datas = [data]
        else:
            failed = False
            # Get all needed infos
            if 'peaks' not in info:
                data = {}
                for k in self.dataframe.columns:
                    data[k] = '-'
                data['Filename'] = file_name
                data['comment'] = "No effective peaks detected"
                new_datas = [data]
                failed = True
            else:
                if 'fit_results' in info:
                    fit_results = info['fit_results']
                    all_S = fit_results['all_S']
                    for side in ['left', 'right']:
                        areas = fit_results[side+'_areas']

                        for i in range(len(areas)):
                            data = {
                                "Filename": file_name,
                                "Side": side,
                                "Distance From Center": all_S[i],
                                "Area" : areas[i]
                            }
                            if i == 0:
                                first_row = data
                            new_datas.append(data)

                        # Write data to the first row of each side
                        first_row.update({
                            "Sigma C": fit_results[side+'_sigmac'],
                            "Sigma D": fit_results[side+'_sigmad'],
                            "Sigma S": fit_results[side+'_sigmas'],
                            "I11/I10": fit_results[side+'_ratio'],
                        })
                        if fit_results["model"] == 'Voigt':
                            first_row["gamma"] = fit_results[side+'_gamma']
                        if fit_results["isSkeletal"]:
                            first_row["Z line"] = fit_results[side+'_zline']
                            first_row["Sigma Z"] = fit_results[side+'_sigmaz']
                            first_row["Iz"] = fit_results[side+'_intz']
                            if fit_results["model"] == 'Voigt':
                                first_row["gamma Z"] = fit_results[side+'_gammaz']


                    # Write general information to the first row of image
                    first_row = new_datas[0]

                    if fit_results["fiterror"] > 0.2:
                        first_row['comment'] = 'High Fitting Error'
                        failed = True

                    first_row['S10'] = fit_results['S10']
                    first_row['Model'] = fit_results['model']
                    first_row['Avarage I11/I10 per fiber'] = fit_results['avg_ratio']
                    first_row['Model'] = fit_results['model']
                    first_row['Fitting error'] = fit_results['fiterror']
                    first_row['CenterX'] = fit_results['centerX']

                    if 'd10' in fit_results.keys():
                        first_row['d10'] = fit_results['d10']
                    else:
                        first_row['d10'] = '-'
                else:
                    data = {}
                    for k in self.dataframe.columns:
                        data[k] = '-'
                    data['Filename'] = file_name
                    data['comment'] = "Model cannot be fit"
                    new_datas = [data]
                    failed = True

            if failed:
                self.failedcases.add(file_name)
            elif file_name in self.failedcases:
                self.failedcases.remove(file_name)

        self.dataframe = self.dataframe.append(new_datas, ignore_index=True)
        self.dataframe.reset_index()
        self.dataframe.to_csv(self.filename, index=False, columns=self.colnames) # Write to csv file

        # Write all failed cases to failed cases file
        f = open(self.failedcasesfile, 'w')
        f.write("\n".join(list(self.failedcases)))
        f.close()

    def removeData(self, file_name):
        """
        Remove data from dataframe
        :param file_name: (str)
        :return:
        """
        self.dataframe = self.dataframe[self.dataframe["Filename"] != file_name]



#cython: boundscheck=False, wraparound=False, nonecheck=False
__author__ = 'Jiranun.J'
import numpy as np
cimport numpy as np
cimport cython
from libc.math cimport sqrt
from libc.math cimport exp

@cython.boundscheck(False)
@cython.wraparound(False)

#
# "ctypedef" assigns a corresponding compile-time type to np.uint8_t. For
# every type in the np module there's a corresponding compile-time
# type with a _t-suffix.
cpdef float distance(int x1, int y1, int x2, int y2):
    return sqrt((x1-x2)**2+(y1-y2)**2)


def get_avg_fold_int16(np.ndarray[np.uint16_t, ndim=3] quadrants, int nQuadrant, int fold_height, int fold_width, double threshold):
    cdef np.ndarray[double, ndim=2] result = np.zeros((fold_height, fold_width))
    cdef Py_ssize_t x, y, n_fold, i
    cdef double sum_val
    cdef np.ndarray[np.uint16_t, ndim=2] fold

    # cdef int end = center_y + box_width
    if nQuadrant > 0:
        for x in range(fold_width):
            for y in range(fold_height):
                sum_val = 0.0
                n_fold = 0
                for i in range(nQuadrant):
                    fold = quadrants[i]
                    if fold[y,x] > threshold :
                        sum_val += fold[y,x]
                        n_fold += 1
                if n_fold == 0 :
                    result[y,x] = 0
                else:
                    result[y,x] = sum_val/n_fold
    return result

def get_avg_fold_int8(np.ndarray[np.uint8_t, ndim=3] quadrants, int nQuadrant, int fold_height, int fold_width, int threshold):
    cdef np.ndarray[double, ndim=2] result = np.zeros((fold_height, fold_width))
    cdef Py_ssize_t x, y, n_fold, i
    cdef double sum_val
    cdef np.ndarray[np.uint8_t, ndim=2] fold

    # cdef int end = center_y + box_width
    if nQuadrant > 0:
        for x in range(fold_width):
            for y in range(fold_height):
                sum_val = 0.0
                n_fold = 0
                for i in range(nQuadrant):
                    fold = quadrants[i]
                    if fold[y,x] > threshold :
                        sum_val += fold[y,x]
                        n_fold += 1
                if n_fold == 0 :
                    result[y,x] = 0
                else:
                    result[y,x] = sum_val/n_fold
    return result

def get_avg_fold_float32(np.ndarray[np.float32_t, ndim=3] quadrants, int nQuadrant, int fold_height, int fold_width, np.float32_t threshold):
    cdef np.ndarray[double, ndim=2] result = np.zeros((fold_height, fold_width))
    cdef Py_ssize_t x, y, n_fold, i
    cdef double sum_val
    cdef np.ndarray[np.float32_t, ndim=2] fold

    # cdef int end = center_y + box_width
    if nQuadrant > 0:
        for x in range(fold_width):
            for y in range(fold_height):
                sum_val = 0.0
                n_fold = 0
                for i in range(nQuadrant):
                    fold = quadrants[i]
                    if fold[y,x] > threshold :
                        sum_val += fold[y,x]
                        n_fold += 1
                if n_fold == 0 :
                    result[y,x] = 0
                else:
                    result[y,x] = sum_val/n_fold
    return result

#
# def apply_radial_bgsub_int8(np.ndarray[np.uint8_t, ndim=2] img, np.ndarray[np.uint8_t, ndim=2] smooth_img, int center_x, int center_y, int radius, int threshold):
#     cdef int img_height = img.shape[0]
#     cdef int img_width = img.shape[1]
#     cdef np.ndarray[np.uint8_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint8)
#     cdef np.ndarray[np.uint8_t, ndim=2] radial_bgsub = np.full((radius,1), np.max(img), dtype = np.uint8)
#     cdef Py_ssize_t x, y
#     cdef int r
#     cdef np.uint8_t bgsub
#     cdef np.uint8_t val, min
#     cdef np.uint8_t zero = 0
#
#     for x in range(img_width):
#         for y in range(img_height):
#             val = smooth_img[y,x]
#             if val >= threshold:
#                 r = int(distance( x, y, center_x, center_y))
#                 min = radial_bgsub[r]
#                 if min > val:
#                     radial_bgsub[r] = val
#
#     for x in range(img_width):
#         for y in range(img_height):
#             val = img[y,x]
#             if val >= threshold:
#                 bgsub = 0
#                 r = int(distance( x, y, center_x, center_y))
#                 if radius > r:
#                     bgsub = radial_bgsub[r]
#
#                 if bgsub <= val:
#                     result[y,x] = val - bgsub
#                 else:
#                     result[y,x] = zero
#             else:
#                result[y,x] = val
#
#     return result
#
#
# def apply_radial_bgsub_int16(np.ndarray[np.uint16_t, ndim=2] img, np.ndarray[np.uint16_t, ndim=2] smooth_img, int center_x, int center_y, int radius, int threshold):
#     cdef int img_height = img.shape[0]
#     cdef int img_width = img.shape[1]
#     cdef np.ndarray[np.uint16_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint16)
#     cdef np.ndarray[np.uint16_t, ndim=2] radial_bgsub = np.full((radius,1), np.max(img), dtype = np.uint16)
#     cdef Py_ssize_t x, y
#     cdef int r
#     cdef np.uint16_t bgsub
#     cdef np.uint16_t val, min
#     cdef np.uint16_t zero = 0
#
#     for x in range(img_width):
#         for y in range(img_height):
#             val = smooth_img[y,x]
#             if val >= threshold:
#                 r = int(distance( x, y, center_x, center_y))
#                 min = radial_bgsub[r]
#                 if min > val:
#                     radial_bgsub[r] = val
#
#     for x in range(img_width):
#         for y in range(img_height):
#             val = img[y,x]
#             if val >= threshold:
#                 bgsub = 0
#                 r = int(distance( x, y, center_x, center_y))
#                 if radius > r:
#                     bgsub = radial_bgsub[r]
#
#                 if bgsub <= val:
#                     result[y,x] = val - bgsub
#                 else:
#                     result[y,x] = zero
#             else:
#                result[y,x] = val
#
#     return result


def apply_radial_bgsub_float32(np.ndarray[np.float32_t, ndim=2] img, np.ndarray[np.float32_t, ndim=2] smooth_img, int center_x, int center_y, int radius, int threshold):
    cdef int img_height = img.shape[0]
    cdef int img_width = img.shape[1]
    cdef np.ndarray[np.float32_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.float32)
    cdef np.ndarray[np.float32_t, ndim=2] radial_bgsub = np.full((radius,1), np.max(img), dtype = np.float32)
    cdef Py_ssize_t x, y
    cdef int r
    cdef np.float32_t bgsub
    cdef np.float32_t val, min
    cdef np.float32_t zero = 0.0

    for x in range(img_width):
        for y in range(img_height):
            val = smooth_img[y,x]
            if val >= threshold:
                r = int(distance( x, y, center_x, center_y))
                min = radial_bgsub[r]
                if min > val:
                    radial_bgsub[r] = val

    for x in range(img_width):
        for y in range(img_height):
            val = img[y,x]
            if val >= threshold:
                bgsub = 0.0
                r = int(distance( x, y, center_x, center_y))
                if radius > r:
                    bgsub = radial_bgsub[r]

                if bgsub <= val:
                    result[y,x] = val - bgsub
                else:
                    result[y,x] = zero
            else:
               result[y,x] = val

    return result


cdef float sigmoid(float k, int x0, float x):
    return 1.0/(1.0+exp(-k*(x-x0)))


def combine_bgsub_float32(np.ndarray[np.float32_t, ndim=2] img1, np.ndarray[np.float32_t, ndim=2] img2, int center_x, int center_y, float sigmoid_k, int radius):
    cdef int img_height = img1.shape[0]
    cdef int img_width = img1.shape[1]
    cdef np.ndarray[np.float32_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.float32)
    cdef int x, y
    cdef np.float32_t r
    cdef np.float32_t tophat_ratio, radial_ratio
    cdef np.float32_t tophat_val, radial_val, th_tmp

    for x in range(img_width):
        for y in range(img_height):
            r = distance( x, y, center_x, center_y)
            tophat_ratio = sigmoid(sigmoid_k, radius, r)
            radial_ratio = 1.0 - tophat_ratio
            tophat_val = tophat_ratio * img2[y,x]
            radial_val = radial_ratio * img1[y,x]
            result[y,x] = tophat_val+radial_val
    return result

# def combine_bgsub_uint16(np.ndarray[np.uint16_t, ndim=2] tophat_img, np.ndarray[np.uint16_t, ndim=2] radial_img, int center_x, int center_y, float sigmoid_k, int radius, int threshold):
#     cdef int img_height = tophat_img.shape[0]
#     cdef int img_width = tophat_img.shape[1]
#     cdef np.ndarray[np.uint16_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint16)
#     cdef Py_ssize_t x, y
#     cdef int r
#     cdef float tophat_ratio, radial_ratio
#     cdef np.uint16_t tophat_val, radial_val, th_tmp
#
#     for x in range(img_width):
#         for y in range(img_height):
#             th_tmp = tophat_img[y,x]
#             if th_tmp >= threshold:
#                 r = int(distance( x, y, center_x, center_y))
#                 tophat_ratio = sigmoid(sigmoid_k, radius, r)
#                 radial_ratio = 1.0 - tophat_ratio
#                 tophat_val = int(tophat_ratio * th_tmp)
#                 radial_val = int(radial_ratio * radial_img[y,x])
#                 result[y,x] = tophat_val+radial_val
#             else:
#                 result[y,x] = th_tmp
#
#     return result
#
# def combine_bgsub_uint8(np.ndarray[np.uint8_t, ndim=2] tophat_img, np.ndarray[np.uint8_t, ndim=2] radial_img, int center_x, int center_y, float sigmoid_k, int radius, int threshold):
#     cdef int img_height = tophat_img.shape[0]
#     cdef int img_width = tophat_img.shape[1]
#     cdef np.ndarray[np.uint8_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint8)
#     cdef Py_ssize_t x, y
#     cdef int r
#     cdef float tophat_ratio, radial_ratio
#     cdef np.uint8_t tophat_val, radial_val, th_tmp
#
#     for x in range(img_width):
#         for y in range(img_height):
#             th_tmp = tophat_img[y,x]
#             if th_tmp >= threshold:
#                 r = int(distance( x, y, center_x, center_y))
#                 tophat_ratio = sigmoid(sigmoid_k, radius, r)
#                 radial_ratio = 1.0 - tophat_ratio
#                 tophat_val = int(tophat_ratio * th_tmp)
#                 radial_val = int(radial_ratio * radial_img[y,x])
#                 result[y,x] = tophat_val+radial_val
#             else:
#                 result[y,x] = th_tmp
#
#     return result


def getCirSubtr_f32(np.ndarray[np.float32_t, ndim=2] img, np.ndarray[np.float32_t, ndim=1] subtr_hist, int radius):
    cdef int img_height = img.shape[0]
    cdef int img_width = img.shape[1]
    cdef np.ndarray[np.float32_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.float32)
    cdef Py_ssize_t x, y
    cdef int r
    cdef np.float32_t bgsub
    cdef np.float32_t val, min
    cdef np.float32_t zero = 0.0

    for x in range(img_width):
        for y in range(img_height):
            val = img[y,x]
            bgsub = 0.0
            r = int(distance( x, y, img_width, img_height))
            if radius > r:
                bgsub = subtr_hist[r]

            result[y, x] = val - bgsub
            # if bgsub <= val:
            #     result[y,x] = val - bgsub
            # else:
            #     result[y,x] = zero
    return result

def getCirSubtr_i16(np.ndarray[np.uint16_t, ndim=2] img, np.ndarray[np.uint16_t, ndim=1] subtr_hist, int radius):
    cdef int img_height = img.shape[0]
    cdef int img_width = img.shape[1]
    cdef np.ndarray[np.uint16_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint16)
    cdef Py_ssize_t x, y
    cdef int r
    cdef np.uint16_t bgsub
    cdef np.uint16_t val, min
    cdef np.uint16_t zero = 0

    for x in range(img_width):
        for y in range(img_height):
            val = img[y,x]
            bgsub = 0
            r = int(distance( x, y, img_width, img_height))
            if radius > r:
                bgsub = subtr_hist[r]

            # result[y, x] = val - bgsub
            if bgsub <= val:
                result[y,x] = val - bgsub
            else:
                result[y,x] = zero
    return result

def getCirSubtr_i8(np.ndarray[np.uint8_t, ndim=2] img, np.ndarray[np.uint8_t, ndim=1] subtr_hist, int radius):
    cdef int img_height = img.shape[0]
    cdef int img_width = img.shape[1]
    cdef np.ndarray[np.uint8_t, ndim=2] result = np.zeros((img_height, img_width), dtype = np.uint8)
    cdef Py_ssize_t x, y
    cdef int r
    cdef np.uint8_t bgsub
    cdef np.uint8_t val, min
    cdef np.uint8_t zero = 0

    for x in range(img_width):
        for y in range(img_height):
            val = img[y,x]
            bgsub = 0
            r = int(distance( x, y, img_width, img_height))
            if radius > r:
                bgsub = subtr_hist[r]

            # result[y, x] = val - bgsub
            if bgsub <= val:
                result[y,x] = val - bgsub
            else:
                result[y,x] = zero
    return result

from libc.math cimport abs, M_PI
from libc.math cimport round
from libc.math cimport atan
from libc.math cimport floor
from libc.math cimport ceil
# from numpy.math cimport rad2deg

@cython.cdivision(True)
def make2DConvexhullBG(np.ndarray[np.float32_t, ndim=2] pchipLines, int width, int height, int centerX, int centerY, int rmin, int rmax):
    cdef np.ndarray[np.float32_t, ndim=2] backgound = np.zeros((height, width), dtype = np.float32)
    cdef int x, y
    cdef int irad, irad_floor, irad_ceil
    cdef np.float32_t deg, slope, floor_deg, alpha, ceil_deg, beta, round_deg, deltay, deltax, subv, tmpdeg
    cdef np.float32_t rad, ceil_rad, floor_rad, alpha_rad, beta_rad
    cdef np.float32_t zero = 0.0
    cdef int intzero = 0
    cdef int pos1, pos2
    cdef int ifloor, iceil, ideg

    for x in range(width):
        for y in range(height):
            deltax = float(abs(x - centerX))
            if deltax == zero:
                deg = 90.0
            else:
                deltay = float(abs(y - centerY))
                slope = deltay / deltax
                deg = atan(slope)*180.0/M_PI

            rad = distance(centerX, centerY, x, y)
            irad = int(round(rad))
            ceil_rad = ceil(rad)
            floor_rad = floor(rad)
            irad_ceil = int(ceil_rad)
            irad_floor = int(floor_rad)
            alpha_rad = 1. - (rad-floor_rad)
            beta_rad = 1. - (ceil_rad-rad)

            round_deg = round(deg)

            if round_deg < deg:
                ceil_deg = round_deg + 0.5
                floor_deg = round_deg
            elif round_deg > deg:
                ceil_deg = round_deg
                floor_deg = round_deg - 0.5
            else:
                ceil_deg = round_deg
                floor_deg = round_deg

            # alpha = 1. - ((deg - floor_deg) * 2.0)
            # beta = 1. - ((ceil_deg - deg) * 2.0)

            ifloor = int(floor_deg / 0.5)
            iceil = int(ceil_deg / 0.5)
            tmpdeg = deg/0.5
            alpha = 1. - (tmpdeg - floor_deg / 0.5)
            beta = 1. - (ceil_deg / 0.5 - tmpdeg)
            ideg = int(round(deg/0.5))

            if irad_ceil <= rmax and irad_floor >= rmin:
                if alpha == 1. and beta == 1.:
                    if alpha_rad == 1. and beta_rad == 1.:
                        pos1 = irad - rmin
                        backgound[y,x] = pchipLines[ideg, pos1]
                    else:
                        pos1 = irad_floor - rmin
                        pos2 = irad_ceil - rmin
                        backgound[y, x] = alpha_rad * pchipLines[ideg, pos1] + beta_rad * pchipLines[ideg, pos2]
                else:
                    if alpha_rad == 1. and beta_rad == 1.:
                        pos1 = irad - rmin
                        backgound[y,x] = alpha * pchipLines[ifloor, pos1] + beta * pchipLines[iceil, pos1]
                    else:
                        pos1 = irad_floor - rmin
                        pos2 = irad_ceil - rmin
                        backgound[y, x] = alpha * (alpha_rad * pchipLines[ifloor, pos1] + beta_rad * pchipLines[ifloor, pos2]) \
                                          + beta * (alpha_rad * pchipLines[iceil, pos1] + beta_rad * pchipLines[iceil, pos2])
    return backgound

@cython.cdivision(True)
def make2DConvexhullBG2(np.ndarray[np.float32_t, ndim=2] pchipLines, int width, int height, int centerX, int centerY, int rmin, int rmax):
    cdef np.ndarray[np.float32_t, ndim=2] backgound = np.zeros((height, width), dtype = np.float32)
    cdef int x, y
    cdef int irad, irad_floor, irad_ceil
    cdef np.float32_t deg, slope, floor_deg, alpha, ceil_deg, beta, round_deg, deltay, deltax, subv, tmpdeg
    cdef np.float32_t rad, ceil_rad, floor_rad, alpha_rad, beta_rad
    cdef np.float32_t zero = 0.0
    cdef int intzero = 0
    cdef int pos1, pos2
    cdef int ifloor, iceil, ideg

    for x in range(width):
        for y in range(height):

            rad = distance(centerX, centerY, x, y)
            irad = int(round(rad))
            ceil_rad = ceil(rad)
            floor_rad = floor(rad)
            irad_ceil = int(ceil_rad)
            irad_floor = int(floor_rad)

            if irad_floor == irad_ceil:
                beta_rad = 0.5
                alpha_rad = 0.5
            else:
                alpha_rad = 1. - (rad - floor_rad)
                beta_rad = 1. - (ceil_rad - rad)

            deltax = float(abs(x - centerX))
            if deltax == zero:
                deg = 90.0
            else:
                deltay = float(abs(y - centerY))
                slope = deltay / deltax
                deg = atan(slope)*180.0/M_PI

            ideg = int(round(deg))
            floor_deg = floor(deg)
            ceil_deg = ceil(deg)
            ifloor = int(floor_deg)
            iceil = int(ceil_deg)

            if ifloor == iceil:
                alpha = 0.5
                beta = 0.5
            else:
                alpha = 1. - (deg - floor_deg)
                beta = 1. - (ceil_deg - deg)

            if irad_ceil < rmax and irad_floor >= rmin:
                pos1 = irad_floor - rmin
                pos2 = irad_ceil - rmin
                backgound[y, x] = alpha * (alpha_rad * pchipLines[ifloor, pos1] + beta_rad * pchipLines[ifloor, pos2]) \
                                  + beta * (alpha_rad * pchipLines[iceil, pos1] + beta_rad * pchipLines[iceil, pos2])
    return backgound


@cython.cdivision(True)
def createAngularBG(int width, int height, np.ndarray[np.float32_t, ndim=2] subtr, int nBins):
    cdef np.ndarray[np.float32_t, ndim=2] backgound = np.zeros((height, width), dtype = np.float32)
    cdef int x,y, irad, ideg
    cdef int centerX = width - 1
    cdef int centerY = height - 1
    cdef np.float32_t rad, deg, slope, deltax, deltay, bgsub, fbin, floor_bin, ceil_bin, alpha, beta, floor_rad, ceil_rad, alpha_rad, beta_rad
    cdef np.float32_t theta_size = 90./nBins
    cdef int ifloor, iceil, ibin, ifloor_rad, iceil_rad

    for x in range(0, width):
        for y in range(0, height):
            rad = distance(centerX, centerY, x, y)
            floor_rad = floor(rad)
            ceil_rad = ceil(rad)
            ifloor_rad = int(floor_rad)
            iceil_rad = int(ceil_rad)

            if ifloor_rad == iceil_rad:
                beta_rad = 0.5
                alpha_rad = 0.5
            else:
                alpha_rad = 1. - (rad - floor_rad)
                beta_rad = 1. - (ceil_rad - rad)

            irad = int(round(rad))

            deltax = float(abs(x - centerX))
            bgsub = 0.0
            if deltax == 0.0:
                deg = 90.0
            else:
                deltay = float(abs(y - centerY))
                slope = deltay / deltax
                deg = atan(slope)*180.0/M_PI

            fbin = 1.*deg/theta_size
            ibin = int(round(fbin))

            if ibin == 0:
                backgound[y, x] = alpha_rad*subtr[ibin, ifloor_rad] + beta_rad*subtr[ibin, iceil_rad]
            elif ibin == nBins:
                backgound[y, x] = alpha_rad*subtr[ibin-1, ifloor_rad] + beta_rad*subtr[ibin-1, iceil_rad]
            else:
                floor_bin = floor(fbin)
                ceil_bin = ceil(fbin)
                alpha = 1. - (fbin - floor_bin)
                beta = 1. - (ceil_bin - fbin)

                if alpha == 1.0 and beta == 1.0:
                    alpha = 0.5
                    beta = 0.5
                    ifloor = int(floor_bin - 1.0)
                    iceil = int(ceil_bin)
                elif alpha > beta :
                    floor_bin = floor_bin - 0.5
                    ceil_bin = ceil_bin - 0.5
                    alpha = 1. - (fbin - floor_bin)
                    beta = 1. - (ceil_bin - fbin)
                    ifloor = int(floor(floor_bin))
                    iceil = int(floor(ceil_bin))
                else:
                    floor_bin = floor_bin + 0.5
                    ceil_bin = ceil_bin + 0.5
                    alpha = 1. - (fbin - floor_bin)
                    beta = 1. - (ceil_bin - fbin)
                    ifloor = int(floor(floor_bin))
                    iceil = int(floor(ceil_bin))

                backgound[y, x] = alpha * (alpha_rad * subtr[ifloor, ifloor_rad] + beta_rad * subtr[ifloor, iceil_rad])+ beta * (alpha_rad * subtr[iceil, ifloor_rad] + beta_rad * subtr[iceil, iceil_rad])

    return backgound

@cython.cdivision(True)
def createCircularlySymBG(int width, int height, np.ndarray[np.float32_t, ndim=1] spline):
    cdef np.ndarray[np.float32_t, ndim=2] backgound = np.zeros((height, width), dtype = np.float32)
    cdef int x, y, ifloor, iceil, irad
    cdef np.float32_t centerX = width - 0.5
    cdef np.float32_t centerY = height - 0.5
    cdef np.float32_t rad, alpha, beta, ffloor, fceil
    cdef np.float32_t fx, fy

    for x in range(width):
        for y in range(height):
            fx = float(x)
            fy = float(y)
            rad = sqrt((fx-centerX)**2+(fy-centerY)**2)
            irad = int(round(rad))
            ffloor = floor(rad)
            fceil = ceil(rad)
            alpha = 1.-(rad-ffloor)
            beta = 1.-(fceil-rad)
            ifloor = int(ffloor)
            iceil = int(fceil)

            if ifloor == iceil:
                alpha = 0.5
                beta = 0.5

            backgound[y, x] = alpha*spline[ifloor] + beta*spline[iceil]

    return backgound

@cython.cdivision(True)
def getCircularDiscreteBackground(np.ndarray[np.float32_t, ndim=2] img, int rmin, float start_p, float end_p, int radial_bin, int nBin, int max_pts):
    cdef int height = img.shape[0]
    cdef int width = img.shape[1]
    cdef np.ndarray[np.float32_t, ndim=1] xs = np.zeros(nBin, dtype = np.float32)
    cdef np.ndarray[np.float32_t, ndim=1] ys = np.zeros(nBin, dtype = np.float32)
    cdef np.ndarray[np.float32_t, ndim=1] all_pts = np.zeros(max_pts, dtype = np.float32)
    cdef np.float32_t centerX = width - 0.5
    cdef np.float32_t centerY = height - 0.5
    cdef int nPoints = 0
    cdef int rad, bin, start_ind, end_ind
    cdef int x, y, i
    cdef np.float32_t float_x, float_y, distance, d1, d2, sumVal

    for bin in range(0, nBin):
        nPoints = 0
        d1 = float(rmin + bin*radial_bin)
        d2 = d1 + float(radial_bin)

        # Get all points in a bin
        for x in range(width):
            float_x = float(x)
            for y in range(height):
                float_y = float(y)
                distance = sqrt((float_x-centerX)**2+(float_y-centerY)**2)
                if d1 <= distance and distance < d2 and nPoints < max_pts:
                    all_pts[nPoints] = img[y, x]
                    nPoints = nPoints+1

        # Sort all pixels
        sorted_pts = all_pts[:nPoints]
        sorted_pts.sort()
        # print sorted_pts

        # Get background value from all points (between percentage of start_p and end_p)
        start_ind = int(round(float(nPoints)*start_p/100))
        end_ind = int(round(float(nPoints)*end_p/100.))
        print nPoints, start_ind, end_ind

        if start_ind < end_ind:
            sumVal = 0.0
            for i in range(start_ind, end_ind):
                sumVal = sumVal + sorted_pts[i]
            ys[bin] = sumVal/float(end_ind-start_ind)
        else:
            ys[bin] = all_pts[start_ind]

        xs[bin] = (d1+d2)/2.

    return xs, ys

@cython.cdivision(True)
def replaceRmin(np.ndarray[np.float32_t, ndim=2] img, int rmin, double val):
    cdef int x, y
    cdef int height = img.shape[0]
    cdef int width = img.shape[1]
    cdef np.float32_t centerX = width
    cdef np.float32_t centerY = height
    cdef np.float32_t distance
    cdef np.float32_t frmin = float(rmin)
    cdef np.ndarray[np.float32_t, ndim=2] result = np.array(img, dtype = np.float32)
    cdef np.float32_t replace_val = float(val)

    for x in range(width-rmin-1, width):
        float_x = float(x)
        for y in range(height-rmin-1, height):
            float_y = float(y)
            distance = sqrt((float_x-centerX)**2+(float_y-centerY)**2)
            if distance <= frmin:
                result[y, x] = replace_val

    return result


import traceback, subprocess, os.path

def mac_except_hook(*exc_info):
    fname = os.path.expanduser('~/Documents/musclex.error.log')
    with open(fname, 'a') as f:
        f.write(''.join(traceback.format_exception(*exc_info)))
        f.close()
    applescript = "'\
tell app \"System Events\" to \
display dialog \"Error details are written to the file: {}.\" \
with title \"Error message\" \
with icon caution \
buttons {{\"OK\"}}'".format(fname)
    # print(applescript)
    subprocess.Popen('osascript -e ' + applescript, shell=True)

handlers = {
    'darwin': mac_except_hook
}


"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""

from ..modules.ScanningDiffraction import *
import pandas as pd

class CP_CSVManager():
    def __init__(self, dir_path):
        self.df_sum = None
        self.df_rings = None
        result_path = fullPath(dir_path, 'cp_results')
        createFolder(result_path)
        self.sum_file = fullPath(result_path, "summary.csv")
        self.rings_file = fullPath(result_path, "rings.csv")
        self.sum_header = ['filename', 'total intensity', 'number of rings']
        self.rings_header = ['filename', 'ring', 'S', 'd', 'peak sigma', 'peak intensity', 'angle', 'angle sigma', 'angle amplitude', 'angle fitting error']
        self.load_all()

    def load_all(self):
        """
        Load summary.csv and rings.csv file and keep data in self.df_sum and self.df_rings
        :return:
        """
        if exists(self.sum_file):
            self.df_sum = pd.read_csv(self.sum_file)
        else:
            self.df_sum = pd.DataFrame(columns=self.sum_header)
        if exists(self.rings_file):
            self.df_rings = pd.read_csv(self.rings_file)
        else:
            self.df_rings = pd.DataFrame(columns=self.rings_header)

    def remove_data(self, filename):
        """
        Remove data from dataframe
        :param file_name: (str)
        :return:
        """
        self.df_sum = self.df_sum[self.df_sum["filename"] != filename]
        self.df_rings = self.df_rings[self.df_rings["filename"] != filename]

    def write_new_data(self, cir_proj):
        """
        Add new data to dataframe, then re-write summary.csv and rings.csv
        :param cir_proj: ScanningDiffraction object with results in its info dict
        :return: -
        """

        file_name = cir_proj.filename
        info = cir_proj.info
        self.remove_data(file_name)

        # Add data to summary.csv
        new_sum_data = {
            'filename' : file_name,
            'total intensity' : info['area']
        }

        if 'model_peaks' in info:
            new_sum_data['number of rings'] = len(info['model_peaks'])

        self.df_sum = self.df_sum.append(new_sum_data, ignore_index = True)
        self.df_sum.reset_index()
        self.df_sum.to_csv(self.sum_file, index=False, columns=self.sum_header)  # Write to csv file

        # Add data to rings.csv
        if 'model_peaks' in info.keys() and len(info['model_peaks']) > 0 and len(info['merged_peaks']) > 0:
            new_datas = []
            nRings = len(info['model_peaks'])
            models = info['ring_models']
            errors = info['ring_errors']
            fit_result = info['fitResult']
            for i in range(nRings):
                if i not in models:
                    continue
                ring = models[i]
                new_data = {'filename':file_name}
                new_data['ring'] = i+1
                new_data['S'] = fit_result['u' + str(i + 1)]
                new_data['peak sigma'] = fit_result['sigmad' + str(i + 1)]
                new_data['peak intensity'] = fit_result['alpha' + str(i + 1)]
                new_data['angle'] = ring['u']
                new_data['angle sigma'] = ring['sigma']
                new_data['angle amplitude'] = ring['alpha']
                new_data['angle fitting error'] = errors[i]
                if 'peak_ds' in info:
                    new_data['d'] = info['peak_ds'][i]
                else:
                    new_data['d'] = '-'
                new_datas.append(new_data)
            self.df_rings = self.df_rings.append(new_datas, ignore_index=True)
        else:
            for k in self.rings_header:
                new_data = {}
                if k == 'filename':
                    new_data[k] = file_name
                else:
                    new_data[k] = '-'
            self.df_rings = self.df_rings.append(new_data, ignore_index = True)

        self.df_rings.reset_index()
        self.df_rings.to_csv(self.rings_file, index=False, columns=self.rings_header)  # Write to csv file



from .pyqt_utils import *
from matplotlib.patches import Ellipse
from matplotlib.collections import PatchCollection
import h5py
from ..utils.file_manager import *
from ..modules.ScanningDiffraction import *
from ..csv_manager import CP_CSVManager
import musclex
import pandas as pd
from .CPImageWindow import CPImageWindow


class HDFBrowser(QDialog):
    """
    Provide options for HDF Browser - select or create one
    """
    def __init__(self, msg, setting_path):
        """
        initial dialog
        :param msg: message which appear in dialog
        :param setting_path: path that HDF file will be saved
        """
        super(HDFBrowser, self).__init__(None)
        self.msg = msg
        self.path = setting_path
        self.status = 0
        self.hdf_file = ""
        self.initUI()
        self.setConnection()

    def initUI(self):
        self.mainLayout = QGridLayout(self)
        self.setWindowTitle("HDF Browser")
        self.label = QLabel(self.msg)

        self.browseButton = QPushButton("Browse a HDF File")
        self.createButton = QPushButton("Create a HDF File")

        self.browseGrp = QGroupBox("Browsed File")
        self.browseLayout = QVBoxLayout(self.browseGrp)
        self.browsedFile = QLabel()
        self.browseLayout.addWidget(self.browsedFile)
        self.browseGrp.setHidden(True)

        self.createGroup = QGroupBox("Create a File")
        self.createLayout = QGridLayout(self.createGroup)
        self.createGroup.setHidden(True)

        self.x_start = QDoubleSpinBox()
        self.x_start.setRange(-100000, 100000)
        self.y_start = QDoubleSpinBox()
        self.y_start.setRange(-100000, 100000)
        self.x_end = QDoubleSpinBox()
        self.x_end.setRange(-100000, 100000)
        self.y_end = QDoubleSpinBox()
        self.y_end.setRange(-100000, 100000)
        self.x_step = QDoubleSpinBox()
        self.x_step.setRange(-100000, 100000)
        self.y_step = QDoubleSpinBox()
        self.y_step.setRange(-100000, 100000)

        self.createLayout.addWidget(QLabel("Start"), 0, 1, 1, 1)
        self.createLayout.addWidget(QLabel("End"), 0, 2, 1, 1)
        self.createLayout.addWidget(QLabel("Step size"), 0, 3, 1, 1)

        self.createLayout.addWidget(QLabel("X"), 1, 0, 1, 1)
        self.createLayout.addWidget(self.x_start, 1, 1, 1, 1)
        self.createLayout.addWidget(self.x_end, 1, 2, 1, 1)
        self.createLayout.addWidget(self.x_step, 1, 3, 1, 1)
        self.createLayout.addWidget(QLabel("Y"), 2, 0, 1, 1)
        self.createLayout.addWidget(self.y_start, 2, 1, 1, 1)
        self.createLayout.addWidget(self.y_end, 2, 2, 1, 1)
        self.createLayout.addWidget(self.y_step, 2, 3, 1, 1)

        self.bottons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)

        self.mainLayout.addWidget(self.label, 0, 0, 1, 2)
        self.mainLayout.addWidget(self.browseButton, 1, 0, 1, 1)
        self.mainLayout.addWidget(self.createButton, 1, 1, 1, 1)
        self.mainLayout.addWidget(self.browseGrp, 2, 0, 1, 2)
        self.mainLayout.addWidget(self.createGroup, 3, 0, 1, 2)
        self.mainLayout.addWidget(self.bottons,4, 0, 1, 2)

    def setConnection(self):
        """
        Set handler for widgets
        """
        self.browseButton.clicked.connect(self.browseClicked)
        self.createButton.clicked.connect(self.createClicked)
        self.bottons.accepted.connect(self.okClicked)
        self.bottons.rejected.connect(self.reject)

    def browseClicked(self):
        """
        Handle when Browse is clicked
        """
        self.hdf_file = getAFile('HDF (*.hdf)')
        if len(self.hdf_file) > 0:
            self.status = 1
            self.browsedFile.setText(self.hdf_file)
            self.browseGrp.setHidden(False)
            self.createGroup.setHidden(True)

    def createClicked(self):
        """
        Handle when Create is clicked
        """
        self.browseGrp.setHidden(True)
        self.createGroup.setHidden(False)
        self.status = 2

    def okClicked(self):
        """
        Handle when OK is clicked
        """
        if self.status == 0:
            errMsg = QMessageBox()
            errMsg.setText('HDF File is not set')
            errMsg.setInformativeText('Please select a HDF file or create a new one')
            errMsg.setStandardButtons(QMessageBox.Ok)
            errMsg.setIcon(QMessageBox.Warning)
            errMsg.exec_()
        elif self.status == 1:
            self.accept()
        elif self.status == 2:
            # Generate HDF file
            self.hdf_file = join(self.path, 'file.hdf')
            x_start = self.x_start.value()
            y_start = self.y_start.value()
            x_end = self.x_end.value()
            y_end = self.y_end.value()
            x_step = self.x_step.value()
            y_step = self.y_step.value()

            x_nStep = int(np.round((x_end - x_start) / x_step) + 1)
            y_nStep = int(np.round((y_end - y_start) / y_step) + 1)

            data = []
            for j in range(0, y_nStep):
                y = y_start + j*y_step
                for i in range(0, x_nStep):
                    x = x_start + i*x_step
                    data.append((x, y))

            hf = h5py.File(self.hdf_file, 'w')
            data_grp = hf.create_group("data")
            data_grp.create_dataset("BL", data=data)

            hf.close()
            self.accept()

class CPBatchWindow(QMainWindow):
    def __init__(self, mainWin=None, dir_path=""):
        # QDialog.__init__(self, parent)
        QWidget.__init__(self)
        self.filePath = dir_path
        self.widgetList = []
        self.update_plot = {'intensity_maps': True,
                            'ds_maps': True,
                            'arange_maps': True,
                            'vector_maps': True,
                            'ellipse_maps': True
                            }
        self.intesityRange = [0, 1, 1, 2]
        self.mainWin = mainWin
        self.color_maps = 'jet'

        ##Batch Mode Params
        self.stopProcess = False
        self.name_dict = {}
        self.intensity_dict = {}
        self.distance_dict = {}
        self.angrange_dict = {}
        self.orientation_dict = {}
        self.fit_dict = {}
        self.fitcd_dict = {}
        self.coord_dict = {}
        self.hdf_filename = ""
        self.hdf_data = np.array([])
        self.xyIntensity = []
        self.xylim = []
        self.batchmodeImg = None
        self.batchmodeImgDetails = None
        self.batchmodeImgFilename = None
        self.updatingUI = False
        self.plots = {}

        self.vec_UV = []
        self.vec_quiver = None

        self.csvManager = CP_CSVManager(self.filePath)
        self.initUI()
        self.setConnections()
        self.processFolder(self.filePath)

    def initUI(self):
        self.setWindowTitle("Scanning Diffraction v." + musclex.__version__)
        # self.setStyleSheet(getStyleSheet())
        self.centralWidget = QWidget(self)
        self.mainLayout = QGridLayout(self.centralWidget)
        self.setCentralWidget(self.centralWidget)

        #### IMAGE ####
        self.imgFigure = plt.figure()
        self.imgAxes = self.imgFigure.add_subplot(111)
        self.imgCanvas = FigureCanvas(self.imgFigure)
        self.img_maxInt = QDoubleSpinBox()
        self.img_maxInt.setValue(1)
        self.img_maxInt.setKeyboardTracking(False)
        self.img_minInt = QDoubleSpinBox()
        self.img_minInt.setValue(0)
        self.img_minInt.setKeyboardTracking(False)
        self.img_minIntLabel = QLabel("Min intensity")
        self.img_maxIntLabel = QLabel("Max intensity")

        self.imgFrame = QFrame()
        self.imgLayout = QGridLayout(self.imgFrame)
        self.imgLayout.addWidget(self.imgCanvas, 0, 0, 1, 4)
        self.imgLayout.addWidget(self.img_minIntLabel, 1, 0, 1, 1)
        self.imgLayout.addWidget(self.img_minInt, 1, 1, 1, 1)
        self.imgLayout.addWidget(self.img_maxIntLabel, 1, 2, 1, 1)
        self.imgLayout.addWidget(self.img_maxInt, 1, 3, 1, 1)
        self.imgFrame.setHidden(True)

        #### BATCHMODE - tabs
        ## BATCHMODE 1 : Total Intensity Map Tab
        self.intensityTab = QWidget()
        self.intensityTab.setContentsMargins(0, 0, 0, 0)
        self.intensityTabLayout = QGridLayout(self.intensityTab)
        self.intensityMapFigure = plt.figure()
        self.intensityMapAxes = self.intensityMapFigure.add_subplot(111)
        self.intensityMapCanvas = FigureCanvas(self.intensityMapFigure)
        self.intensityMapFigure.canvas.mpl_connect('button_press_event', self.plotClicked)
        self.int_maxIntMap = QDoubleSpinBox()
        self.int_maxIntMap.setMinimum(1)
        self.int_maxIntMap.setMaximum(100)
        self.int_maxIntMap.setValue(100.)
        self.int_maxIntMap.setSingleStep(5.0)
        self.int_maxIntMap.setSuffix("%")
        self.int_maxIntMap.setKeyboardTracking(False)

        self.int_minIntMap = QDoubleSpinBox()
        self.int_minIntMap.setMinimum(0)
        self.int_minIntMap.setMaximum(99)
        self.int_minIntMap.setValue(0)
        self.int_minIntMap.setSingleStep(5.0)
        self.int_minIntMap.setSuffix("%")
        self.int_minIntMap.setKeyboardTracking(False)

        self.int_MapIntSettings = QHBoxLayout()
        self.int_MapIntSettings.addWidget(QLabel("Max : "))
        self.int_MapIntSettings.addWidget(self.int_maxIntMap)
        self.int_MapIntSettings.addWidget(QLabel("Min : "))
        self.int_MapIntSettings.addWidget(self.int_minIntMap)
        self.intensityTabLayout.addWidget(self.intensityMapCanvas, 0, 0, 1, 1)
        self.intensityTabLayout.addLayout(self.int_MapIntSettings, 1, 0, 1, 1)

        ## BATCHMODE 2 : Distance btw rings map
        self.distanceMapTab = QWidget()
        self.distanceMapTab.setContentsMargins(0, 0, 0, 0)
        self.distanceMapTabLayout = QGridLayout(self.distanceMapTab)
        self.distanceMapFigure = plt.figure()
        self.distanceMapAxes = self.distanceMapFigure.add_subplot(111)
        self.distanceMapCanvas = FigureCanvas(self.distanceMapFigure)
        self.distanceMapFigure.canvas.mpl_connect('button_press_event', self.plotClicked)
        self.ds_ImgFigure = plt.figure()
        self.ds_ImgCanvas = FigureCanvas(self.ds_ImgFigure)

        self.ds_maxIntMap = QDoubleSpinBox()
        self.ds_maxIntMap.setMinimum(1)
        self.ds_maxIntMap.setMaximum(100)
        self.ds_maxIntMap.setValue(100.)
        self.ds_maxIntMap.setSingleStep(5.0)
        self.ds_maxIntMap.setSuffix("%")
        self.ds_maxIntMap.setKeyboardTracking(False)
        self.ds_minIntMap = QDoubleSpinBox()
        self.ds_minIntMap.setMinimum(0)
        self.ds_minIntMap.setMaximum(99)
        self.ds_minIntMap.setValue(0)
        self.ds_minIntMap.setSingleStep(5.0)
        self.ds_minIntMap.setSuffix("%")
        self.ds_minIntMap.setKeyboardTracking(False)
        self.ds_MapIntSettings = QHBoxLayout()
        self.ds_MapIntSettings.addWidget(QLabel("Max : "))
        self.ds_MapIntSettings.addWidget(self.ds_maxIntMap)
        self.ds_MapIntSettings.addWidget(QLabel("Min : "))
        self.ds_MapIntSettings.addWidget(self.ds_minIntMap)
        self.distanceMapTabLayout.addWidget(self.distanceMapCanvas, 0, 0, 1, 1)
        self.distanceMapTabLayout.addLayout(self.ds_MapIntSettings, 1, 0, 1, 1)

        ## BATCHMODE 3 : Angular range map (degrees)
        self.angularMapTab = QWidget()
        self.angularMapTab.setContentsMargins(0, 0, 0, 0)
        self.angularMapTabLayout = QGridLayout(self.angularMapTab)
        self.angularMapTab = QWidget()
        self.angularMapTab.setContentsMargins(0, 0, 0, 0)
        self.angularMapTabLayout = QGridLayout(self.angularMapTab)
        self.angularMapFigure = plt.figure()
        self.angularMapAxes = self.angularMapFigure.add_subplot(111)
        self.angularMapCanvas = FigureCanvas(self.angularMapFigure)
        self.angularMapFigure.canvas.mpl_connect('button_press_event', self.plotClicked)

        self.ar_maxIntMap = QDoubleSpinBox()
        self.ar_maxIntMap.setMinimum(1)
        self.ar_maxIntMap.setMaximum(100)
        self.ar_maxIntMap.setValue(100.)
        self.ar_maxIntMap.setSingleStep(5.0)
        self.ar_maxIntMap.setSuffix("%")
        self.ar_maxIntMap.setKeyboardTracking(False)
        self.ar_minIntMap = QDoubleSpinBox()
        self.ar_minIntMap.setMinimum(0)
        self.ar_minIntMap.setMaximum(99)
        self.ar_minIntMap.setValue(0)
        self.ar_minIntMap.setSingleStep(5.0)
        self.ar_minIntMap.setSuffix("%")
        self.ar_minIntMap.setKeyboardTracking(False)
        self.ar_MapIntSettings = QHBoxLayout()
        self.ar_MapIntSettings.addWidget(QLabel("Max : "))
        self.ar_MapIntSettings.addWidget(self.ar_maxIntMap)
        self.ar_MapIntSettings.addWidget(QLabel("Min : "))
        self.ar_MapIntSettings.addWidget(self.ar_minIntMap)
        self.angularMapTabLayout.addWidget(self.angularMapCanvas, 0, 0, 1, 1)
        self.angularMapTabLayout.addLayout(self.ar_MapIntSettings, 1, 0, 1, 1)

        ## BATCHMODE 4 : Vector Fields
        self.vectorFieldTab = QWidget()
        self.vectorFieldTab.setContentsMargins(0, 0, 0, 0)
        self.vectorFieldTabLayout = QGridLayout(self.vectorFieldTab)
        self.vectorFieldMapFigure = plt.figure()
        self.vectorFieldMapAxes = self.vectorFieldMapFigure.add_subplot(111)
        self.vectorFieldMapCanvas = FigureCanvas(self.vectorFieldMapFigure)
        self.vectorFieldMapFigure.canvas.mpl_connect('button_press_event', self.plotClicked)
        self.arrowLengthSlider = QSlider()
        self.arrowLengthSlider.setMinimum(5)
        self.arrowLengthSlider.setMaximum(25)
        self.arrowLengthSlider.setValue(5)
        self.arrowLengthSlider.setOrientation(Qt.Horizontal)

        self.vectorFieldTabLayout.addWidget(self.vectorFieldMapCanvas, 0, 0, 1, 2)
        self.vectorFieldTabLayout.addWidget(QLabel("Arrow Length"), 1, 0, 1, 1)
        self.vectorFieldTabLayout.addWidget(self.arrowLengthSlider, 1, 1, 1, 1)

        ## BATCHMODE 5 : Elliptical Presentation
        self.ellipticalTab = QWidget()
        self.ellipticalTab.setContentsMargins(0, 0, 0, 0)
        self.ellipticalTabLayout = QGridLayout(self.ellipticalTab)
        self.ellipticalMapFigure = plt.figure()
        self.ellipticalMapAxes = self.ellipticalMapFigure.add_subplot(111)
        self.ellipticalMapCanvas = FigureCanvas(self.ellipticalMapFigure)
        self.ellipticalMapFigure.canvas.mpl_connect('button_press_event', self.plotClicked)
        self.ellipticalTabLayout.addWidget(self.ellipticalMapCanvas, 0, 0, 1, 1)

        ## tabs
        self.tabWidget = QTabWidget()
        self.tabWidget.setTabPosition(QTabWidget.North)
        self.tabWidget.setDocumentMode(False)
        self.tabWidget.setTabsClosable(False)
        self.tabWidget.setStyleSheet("QTabBar::tab { height: 35px; width: 200px; }")
        self.tabWidget.addTab(self.intensityTab, "Total Intensity Map")
        # self.tabWidget.addTab(self.distanceMapTab, "D-spacing Map")
        self.tabWidget.addTab(self.angularMapTab, "Angular Range Map\n(degrees)")
        self.tabWidget.addTab(self.vectorFieldTab, "Orientation and Intensity\nVector Field")
        self.tabWidget.addTab(self.ellipticalTab, "Elliptical Representation")

        # status bar
        self.statusBar = QStatusBar()
        self.imagePathLabel = QLabel('')
        self.statusLabel = QLabel('')
        self.statusBar.addWidget(QLabel('   '))
        self.statusBar.addWidget(self.imagePathLabel)
        self.progressBar = QProgressBar()
        self.progressBar.setMaximum(100)
        self.progressBar.setMinimum(0)
        self.progressBar.setHidden(True)
        # self.stopButton = QPushButton('STOP')
        # self.stopButton.clicked.connect(self.stopProcessing)
        # self.stopButton.setHidden(True)
        self.moreDetailsButton = QPushButton('More Details')
        self.moreDetailsButton.setHidden(True)
        self.rightBarLayout = QVBoxLayout()
        self.rightBarLayout.addWidget(self.statusLabel)
        self.rightBarLayout.addWidget(self.moreDetailsButton)
        self.rightBarLayout.setAlignment(Qt.AlignRight)
        self.rightBarFrame = QFrame()
        self.rightBarFrame.setLayout(self.rightBarLayout)
        self.statusBar.addPermanentWidget(self.progressBar)
        self.statusBar.addPermanentWidget(self.rightBarFrame)

        self.refreshButton = QPushButton("Refresh")
        self.refreshButton.setFixedHeight(40)

        self.mainLayout.addWidget(self.imgFrame, 0, 0, 1, 1)
        self.mainLayout.addWidget(self.refreshButton, 1, 0, 1, 1)
        self.mainLayout.addWidget(self.tabWidget, 0, 1, 2, 1)
        self.mainLayout.addWidget(self.statusBar, 2, 0, 1, 2)
        self.mainLayout.setRowStretch(0, 10)
        self.mainLayout.setRowStretch(1, 1)
        self.mainLayout.setRowStretch(2, 1)
        # self.mainLayout.setColumnStretch(0, 1)
        # self.mainLayout.setColumnStretch(1, 1)

        self.show()
        self.resize(1000, 1000)

    def setConnections(self):
        self.refreshButton.clicked.connect(self.processBatchmodeResults)

        self.img_maxInt.valueChanged.connect(self.maxIntChanged)
        self.img_minInt.valueChanged.connect(self.minIntChanged)

        self.int_maxIntMap.valueChanged.connect(self.sliderReleased)
        self.int_minIntMap.valueChanged.connect(self.sliderReleased)

        self.ds_maxIntMap.valueChanged.connect(self.sliderReleased)
        self.ds_minIntMap.valueChanged.connect(self.sliderReleased)

        self.ar_maxIntMap.valueChanged.connect(self.sliderReleased)
        self.ar_minIntMap.valueChanged.connect(self.sliderReleased)

        self.arrowLengthSlider.sliderReleased.connect(self.sliderReleased)

        self.tabWidget.currentChanged.connect(self.updateUI)
        self.moreDetailsButton.clicked.connect(self.popupImageDetails)

    def closeEvent(self, ev):
        if self.mainWin is not None:
            self.mainWin.removeWidget(self)

    def removeWidget(self, win):
        if win in self.widgetList:
            idx = self.widgetList.index(win)
            del self.widgetList[idx]

    def sliderReleased(self):
        QApplication.processEvents()
        if self.tabWidget.currentIndex() == 0:
            self.refreshAllTabs()
        elif self.tabWidget.currentIndex() == 1:
            self.update_plot['arange_maps'] = True
        elif self.tabWidget.currentIndex() == 2:
            self.updateVectorFieldArrow()

        self.updateUI()

    def plotClicked(self, event):
        self.imgFrame.setHidden(False)
        self.moreDetailsButton.setHidden(False)

        if len(self.xyIntensity) < 3:
            return

        x = self.xyIntensity[0]
        y = self.xyIntensity[1]
        x_max = len(x)

        if x[0] <= event.xdata <= x[len(x) - 1] and y[0] <= event.ydata <= y[len(y) - 1]:
            # col = 0
            # row = 0
            indexs = list(range(0, len(x)))
            col = min(indexs, key=lambda i: abs(x[i] - event.xdata))

            if (self.tabWidget.currentIndex() == 0 or self.tabWidget.currentIndex() == 1) and event.xdata < x[col]:
                col = max(col - 1, 0)

            if self.tabWidget.currentIndex() == 3:
                col = min(col + 1, len(x) - 1)

            indexs = list(range(0, len(y)))
            row = min(indexs, key=lambda i: abs(y[i] - event.ydata))
            if (self.tabWidget.currentIndex() == 0 or self.tabWidget.currentIndex() == 1) and event.ydata < y[row]:
                row = max(row - 1, 0)

            ind = row * x_max + col + self.init_number

            if ind in self.name_dict:
                img_detail = "Intensity value: " + str(self.intensity_dict[ind])
                # img_detail += "\nD-spacing: " + str(self.distance_dict[ind])
                img_detail += "\nOrientation angle: " + str(self.orientation_dict[ind])
                img_detail += "\nAngular range: " + str(self.angrange_dict[ind])
                filename = self.name_dict[ind]
                full_filename = fullPath(self.filePath, filename)
                self.batchmodeImgDetails = img_detail
                self.batchmodeImgFilename = str(filename)
                self.updateRightStatusBar(filename)

                if exists(full_filename):

                    img = fabio.open(full_filename).data
                    if img is not None:
                        self.batchmodeImg = img
                        self.setMinMaxIntensity(img, self.img_minInt, self.img_maxInt, self.img_minIntLabel,
                                                self.img_maxIntLabel)
                        QApplication.processEvents()
                else:
                    self.batchmodeImg = None

                self.updateImage()
            else:
                self.batchmodeImgDetails = None
                self.batchmodeImgFilename = None
                self.batchmodeImg = None

    def stopProcessing(self):
        self.stopProcess = True

    def popupImageDetails(self):
        if self.batchmodeImg is not None:
            new_image_window = CPImageWindow(self, str(self.batchmodeImgFilename), str(self.filePath))
            self.widgetList.append(new_image_window)
        else:
            if self.batchmodeImgFilename is None:
                errMsg = QMessageBox()
                errMsg.setText('Image has not been selected')
                errMsg.setInformativeText('Please select an image from maps')
                errMsg.setStandardButtons(QMessageBox.Ok)
                errMsg.setIcon(QMessageBox.Warning)
                errMsg.exec_()
            else:
                errMsg = QMessageBox()
                errMsg.setText('Image not found')
                errMsg.setInformativeText(str(self.batchmodeImgFilename) + ' not found.\nPlease select another image')
                errMsg.setStandardButtons(QMessageBox.Ok)
                errMsg.setIcon(QMessageBox.Warning)
                errMsg.exec_()

    def maxIntChanged(self):
        if self.updatingUI:
            return

        if self.img_maxInt.value() < self.img_minInt.value():
            self.img_maxInt.setValue(self.img_minInt.value() + 1)
        else:
            self.updateImage()

    def minIntChanged(self):
        if self.updatingUI:
            return

        if self.img_maxInt.value() < self.img_minInt.value():
            self.img_minInt.setValue(self.img_maxInt.value() - 1)
        else:
            self.updateImage()

    def updateImage(self):
        if self.batchmodeImgDetails is not None:
            ax = self.imgAxes
            ax.cla()
            ax.set_title(self.batchmodeImgFilename)
            ax.set_xlabel(self.batchmodeImgDetails)

            if self.batchmodeImg is not None:
                img = getBGR(get8bitImage(self.batchmodeImg, min=self.img_minInt.value(), max=self.img_maxInt.value()))
                ax.imshow(img)
            else:
                xlim = ax.get_xlim()
                ylim = ax.get_ylim()
                cx = (xlim[0] + xlim[1]) / 2.
                cy = (ylim[0] + ylim[1]) / 2.
                ax.text(cx, cy, "IMAGE NOT FOUND", fontsize=15, horizontalalignment='center')

            self.imgCanvas.draw()

    def mousePressEvent(self, event):
        focused_widget = QApplication.focusWidget()
        if focused_widget != None:
            focused_widget.clearFocus()

    def updateUI(self):
        selected_tab = self.tabWidget.currentIndex()
        if selected_tab == 0:
            self.updateTotalIntenTab()
        elif selected_tab == 1:
            self.updateAngularRangeTab()
        elif selected_tab == 2:
            self.updateVectorFieldTab()
        elif selected_tab == 3:
            self.updateEllipticalTab()
        QApplication.processEvents()

    def updateTotalIntenTab(self):

        if self.update_plot['intensity_maps']:
            if len(self.xyIntensity) < 3:
                return

            x = self.xyIntensity[0]
            y = self.xyIntensity[1]
            intensity = copy.copy(self.xyIntensity[2])

            x_coor, y_coor = np.meshgrid(x, y)

            max_val = intensity.max()
            if self.int_maxIntMap.value() < 100:
                intensity[
                    intensity > self.int_maxIntMap.value() * max_val / 100.] = self.int_maxIntMap.value() * max_val / 100.
            if self.int_minIntMap.value() > 0:
                intensity[
                    intensity < self.int_minIntMap.value() * max_val / 100.] = self.int_minIntMap.value() * max_val / 100.

            ax = self.intensityMapAxes
            ax.cla()
            ax.set_title("Total Intensity Map\n")
            im = ax.pcolormesh(x_coor, y_coor, intensity, cmap=self.color_maps)
            ax.set_xlim(x.min() - self.xylim[0], x.max() + self.xylim[0])
            ax.set_ylim(y.min() - self.xylim[1], y.max() + self.xylim[1])
            # self.intensityMapFigure.colorbar(im)
            self.intensityMapFigure.tight_layout()
            self.intensityMapFigure.savefig(fullPath(self.filePath, 'cp_results/intensity_map.png'))
            self.intensityMapCanvas.draw()
            self.update_plot['intensity_maps'] = False

    def updateDspacingTab(self):

        if self.update_plot['ds_maps']:
            if len(self.xyIntensity) < 3:
                return

            x = self.xyIntensity[0]
            y = self.xyIntensity[1]
            x_max = len(x)

            if 'ds_' not in self.plots.keys():
                distances = [
                    float(self.distance_dict[i]) if i in self.distance_dict and self.distance_dict[i] != '' else 0
                    for i in
                    range(self.init_number, len(self.hdf_data) + self.init_number)]
                distances = np.array([distances[i:i + x_max] for i in range(0, len(self.hdf_data), x_max)])
                self.plots['ds_'] = copy.copy(distances)
            else:
                distances = copy.copy(self.plots['ds_'])

            # z = cv2.blur(z, (4,4))
            x_coor, y_coor = np.meshgrid(x, y)

            max_val = distances.max()
            min_val = distances.min()
            if self.ds_maxIntMap.value() < 100:
                distances[distances > self.ds_maxIntMap.value() * max_val / 100.] = max_val
            if self.ds_minIntMap.value() > 0:
                distances[distances < self.ds_minIntMap.value() * max_val / 100.] = min_val

            ax = self.distanceMapAxes
            ax.cla()
            ax.set_title("D-spacing Map\n")
            im = ax.pcolormesh(x_coor, y_coor, distances, cmap=self.color_maps)
            ax.set_xlim(x.min() - self.xylim[0], x.max() + self.xylim[0])
            ax.set_ylim(y.min() - self.xylim[1], y.max() + self.xylim[1])
            # self.distanceMapFigure.colorbar(im)
            self.distanceMapFigure.tight_layout()
            self.distanceMapFigure.savefig(fullPath(self.filePath, 'cp_results/d_spacing_map.png'))
            self.distanceMapCanvas.draw()
            self.update_plot['ds_maps'] = False

    def updateAngularRangeTab(self):

        if self.update_plot['arange_maps']:
            if len(self.xyIntensity) < 3:
                return

            x = self.xyIntensity[0]
            y = self.xyIntensity[1]
            x_max = len(x)

            if 'ar_' not in self.plots.keys():
                ang_range = [
                    convertRadtoDegrees(float(self.angrange_dict[i])) if i in self.angrange_dict and self.angrange_dict[
                                                                                                         i] != '' else 0
                    for
                    i in range(self.init_number, len(self.hdf_data) + self.init_number)]
                ang_range = [r if 0 < r <= 180 else 0 for r in ang_range]
                ang_range = np.array([ang_range[i:i + x_max] for i in range(0, len(self.hdf_data), x_max)])
                self.plots['ar_'] = copy.copy(ang_range)
            else:
                ang_range = copy.copy(self.plots['ar_'])

            # z = cv2.blur(z, (4,4))
            x_coor, y_coor = np.meshgrid(x, y)

            max_val = ang_range.max()
            min_val = ang_range.min()
            if self.ar_maxIntMap.value() < 100:
                ang_range[ang_range > self.ar_maxIntMap.value() * max_val / 100.] = max_val
            if self.ar_minIntMap.value() > 0:
                ang_range[ang_range < self.ar_minIntMap.value() * max_val / 100.] = min_val

            ax = self.angularMapAxes
            ax.cla()
            ax.set_title("Angular Range Map (Degrees)\n")
            im = ax.pcolormesh(x_coor, y_coor, ang_range, cmap=self.color_maps)
            ax.set_xlim(x.min() - self.xylim[0], x.max() + self.xylim[0])
            ax.set_ylim(y.min() - self.xylim[1], y.max() + self.xylim[1])
            # self.angularMapFigure.colorbar(im)
            self.angularMapFigure.tight_layout()
            self.angularMapFigure.savefig(fullPath(self.filePath, 'cp_results/angular_range_map.png'))
            self.angularMapCanvas.draw()
            self.update_plot['arange_maps'] = False

    def updateVectorFieldTab(self):

        if self.update_plot['vector_maps']:
            if len(self.xyIntensity) < 3:
                return

            x = self.xyIntensity[0]
            y = self.xyIntensity[1]
            x_max = len(x)
            intensity = self.xyIntensity[2]

            orientation = np.array(
                [float(self.orientation_dict[i]) if i in self.orientation_dict and self.orientation_dict[i] != '' else 0
                 for i in
                 range(self.init_number, len(self.hdf_data) + self.init_number)])
            orientation = [np.pi - ang for ang in orientation]
            # orientation = [np.pi - ang if ang < np.pi else ang - np.pi for ang in orientation]
            orientation = np.array([orientation[i:i + x_max] for i in range(0, len(self.hdf_data), x_max)])

            U = np.cos(orientation)
            V = np.sin(orientation)
            int_display = copy.copy(intensity)

            max_val = int_display.max()
            if self.int_maxIntMap.value() < 100:
                int_display[
                    int_display > self.int_maxIntMap.value() * max_val / 100.] = self.int_maxIntMap.value() * max_val / 100.
            if self.int_minIntMap.value() > 0:
                intensity[
                    int_display < self.int_minIntMap.value() * max_val / 100.] = self.int_minIntMap.value() * max_val / 100.

            speed = int_display / intensity.max()
            UN = U * speed
            VN = V * speed
            self.vec_UV = [U, V]

            ax = self.vectorFieldMapAxes
            ax.cla()
            ax.set_facecolor('black')
            ax.set_title("Orientation (direction) and Intensity (height and color) Vector Field")
            self.vec_quiver = ax.quiver(x, y, UN, VN,  # data
                                        int_display,  # colour the arrows based on this array
                                        cmap=self.color_maps,  # colour map
                                        headlength=7, headwidth=4)

            ax.set_xlim(x.min() - self.xylim[0], x.max() + self.xylim[0])
            ax.set_ylim(y.min() - self.xylim[1], y.max() + self.xylim[1])
            ax.set_aspect('auto')

            # self.vectorFieldMapFigure.colorbar(self.vec_quiver)
            self.vectorFieldMapFigure.tight_layout()
            self.vectorFieldMapFigure.savefig(fullPath(self.filePath, 'cp_results/vector_field.png'))
            self.vectorFieldMapCanvas.draw()

            if self.arrowLengthSlider.value() > 5:
                self.updateVectorFieldArrow()

            self.update_plot['vector_maps'] = False

    def updateVectorFieldArrow(self):
        if len(self.xyIntensity) < 3 or len(self.vec_UV) < 2 or self.vec_quiver is None:
            return

        intensity = self.xyIntensity[2]
        U = self.vec_UV[0]
        V = self.vec_UV[1]
        int_display = copy.copy(intensity)

        max_val = int_display.max()
        if self.int_maxIntMap.value() < 100:
            int_display[
                int_display > self.int_maxIntMap.value() * max_val / 100.] = self.int_maxIntMap.value() * max_val / 100.
        if self.int_minIntMap.value() > 0:
            int_display[
                int_display < self.int_minIntMap.value() * max_val / 100.] = self.int_minIntMap.value() * max_val / 100.

        speed = int_display / intensity.max() * (self.arrowLengthSlider.value() / 5.)
        UN = U * speed
        VN = V * speed
        self.vec_quiver.set_UVC(UN, VN)
        self.vectorFieldMapCanvas.draw_idle()
        self.vectorFieldMapFigure.savefig(fullPath(self.filePath, 'cp_results/vector_field.png'))

    def updateEllipticalTab(self):

        if self.update_plot['ellipse_maps']:

            if len(self.xyIntensity) < 3:
                return

            x = self.xyIntensity[0]
            y = self.xyIntensity[1]

            centers = [(x[i], y[j]) for j in range(len(y)) for i in range(len(x))]
            ranges = [toFloat(self.angrange_dict[i]) if i in self.angrange_dict.keys() else 0. for i in
                      range(self.init_number, len(self.hdf_data) + self.init_number)]
            max_width = max(ranges)
            widths = [toFloat(self.angrange_dict[i]) / max_width if i in self.angrange_dict.keys() and 0 < toFloat(
                self.angrange_dict[i]) / max_width else max_width / 2. for i in
                      range(self.init_number, len(self.hdf_data) + self.init_number)]

            int_display = np.array(list(self.intensity_dict.values()))
            max_val = int_display.max()
            if self.int_maxIntMap.value() < 100:
                int_display[
                    int_display > self.int_maxIntMap.value() * max_val / 100.] = self.int_maxIntMap.value() * max_val / 100.
            if self.int_minIntMap.value() > 0:
                int_display[
                    int_display < self.int_minIntMap.value() * max_val / 100.] = self.int_minIntMap.value() * max_val / 100.

            ax = self.ellipticalMapAxes
            ax.cla()
            ax.set_title("Elliptical Representation of Orientation (direction) and Angle Range (width)")
            patches = []
            colors = []
            for i in range(len(self.hdf_data)):

                if ranges[i] == 0:
                    e = Ellipse(xy=centers[i - self.init_number], width=self.xylim[0]/5., height=self.xylim[0]/5.)
                else:
                    e = Ellipse(xy=centers[i - self.init_number], width= self.xylim[0] * widths[i], height=self.xylim[0],
                                angle=convertRadtoDegreesEllipse(
                                    np.pi - toFloat(self.orientation_dict[i + self.init_number])))
                patches.append(e)
                # colors.append(self.intensity_dict[i + self.init_number])
                if i < len(int_display):
                    colors.append(int_display[i])
                else:
                    colors.append(0)

            p = PatchCollection(patches, cmap=self.color_maps)
            p.set_array(np.array(colors))
            ax.add_collection(p)
            ax.set_facecolor('black')
            ax.set_xlim(x.min() - self.xylim[0], x.max() + self.xylim[0])
            ax.set_ylim(y.min() - self.xylim[1], y.max() + self.xylim[1])
            ax.set_aspect('auto')

            # self.ellipticalMapFigure.colorbar(p)
            self.ellipticalMapFigure.tight_layout()
            self.ellipticalMapFigure.savefig(fullPath(self.filePath, 'cp_results/direction_width.png'))
            self.ellipticalMapCanvas.draw()
            self.update_plot['ellipse_maps'] = False

    def browseHDF(self, dir_path, hdfList=[]):
        hdf_filename = ""
        path = join(dir_path, 'settings')
        createFolder(path)
        hdf_cache = join(path, 'hdf.info')

        if exists(hdf_cache):
            hdf_filename = pickle.load(open(hdf_cache, "rb"))

        if len(hdf_filename) == 0 or not exists(hdf_filename):
            if len(hdfList) == 1:
                hdf_filename = join(dir_path, hdfList[0])
            else:
                if len(hdfList) == 0:
                    dlg = HDFBrowser('No HDF file detected.\nPlease select an HDF file to process or create a new one.', path)
                else:
                    dlg = HDFBrowser('There are more than one HDF file detected. \nPlease select an HDF file to process or create a new one.', path)

                ret = dlg.exec_()
                if ret == 0:
                    hdf_filename = ""
                else:
                    hdf_filename = dlg.hdf_file

        if hdf_filename != "":
            pickle.dump(hdf_filename, open(hdf_cache, "wb"))
            self.hdf_filename = str(hdf_filename)
            self.processBatchmodeResults()
        else:
            self.close()

    def processBatchmodeResults(self):
        QApplication.setOverrideCursor(Qt.WaitCursor)
        dir_path = self.filePath
        self.csvManager.load_all()
        hdf_filename = self.hdf_filename
        csv_filename = self.csvManager.sum_file
        self.updateStatusBar(text='Dir : ' + dir_path + '\nHDF : ' + hdf_filename + '\nCSV : ' + csv_filename)
        df_sum = self.csvManager.df_sum.copy()
        df_sum = df_sum.sort_values(['filename'], ascending=True)
        df_rings = self.csvManager.df_rings

        # Read intensity from csv to organize the info given
        self.name_dict = {}
        self.intensity_dict = {}
        self.distance_dict = {}
        self.angrange_dict = {}
        self.orientation_dict = {}
        self.fit_dict = {}
        self.fitcd_dict = {}

        for i, row in df_sum.iterrows():
            filename = str(row['filename'])
            start_ind = filename.rfind('_')
            end_ind = filename.rfind('.')
            index = int(row['filename'][start_ind + 1:end_ind])
            self.name_dict[index] = row['filename']
            self.intensity_dict[index] = row['total intensity']

            # Add ring model if its error < 1. and sigma < 1. (prevent uniform ring)
            all_rings = df_rings[df_rings['filename'] == filename]
            if len(all_rings) > 0:
                all_rings = all_rings.sort_values(['angle fitting error'], ascending=True)
                best_ring = all_rings.iloc[0]
                good_model = float(best_ring['angle fitting error']) < 1. and best_ring['angle sigma'] < 1.
                self.orientation_dict[index] = best_ring['angle'] if pd.notnull(
                    best_ring['angle']) and good_model else 0
                self.angrange_dict[index] = best_ring['angle sigma'] if pd.notnull(
                    best_ring['angle sigma']) and good_model else 0
                self.distance_dict[index] = 0
            else:
                self.orientation_dict[index] = 0
                self.angrange_dict[index] = 0
                self.distance_dict[index] = 0

        self.init_number = min(self.name_dict.keys())

        # Read hdf5 file to get the coordinates and image shape
        hf = h5py.File(hdf_filename, 'r')
        data = hf.get('data').get('BL')
        self.hdf_data = np.array(data)
        self.coord_dict = {}

        for i in range(self.init_number, len(self.hdf_data) + self.init_number):
            self.coord_dict[i] = (self.hdf_data[i - self.init_number][0], self.hdf_data[i - self.init_number][1])

        nCols = 0
        for i in range(self.init_number + 1, len(self.hdf_data) + self.init_number):
            if abs(self.coord_dict[i][1] - self.coord_dict[i - 1][1]) != 0:
                nCols = i - self.init_number
                break

        nRows = len(self.hdf_data) / nCols
        all_xs = np.reshape(np.array([v[0] for k, v in self.coord_dict.items()]), (nRows, nCols))
        all_ys = np.reshape(np.array([v[1] for k, v in self.coord_dict.items()]), (nRows, nCols))

        x = np.mean(all_xs, axis=0)
        y = np.mean(all_ys, axis=1)

        # Check if any error on coordinates
        x_grad = abs(x[1] - x[0])
        y_grad = abs(y[1] - y[0])

        # Plot heatmap for intensity
        z = [float(self.intensity_dict[i]) if i in self.intensity_dict else 0 for i in
             range(self.init_number, len(self.hdf_data) + self.init_number)]
        # z = np.array([z[i:i + x_max] for i in range(0, , x_max)])
        # z = cv2.blur(z, (4,4))
        # intensity = np.array(z)
        intensity = np.reshape(z, (len(y), len(x)))

        self.xyIntensity = [x, y, intensity]
        self.xylim = [x_grad, y_grad]
        self.refreshAllTabs()
        self.updateImage()
        self.updateUI()
        QApplication.restoreOverrideCursor()

    def refreshAllTabs(self):
        # Set all update flags to True
        self.update_plot = {'intensity_maps': True,
                            'ds_maps': True,
                            'arange_maps': True,
                            'vector_maps': True,
                            'ellipse_maps': True
                            }

    def processFolder(self, dir_path):
        imgList, hdfList = getFilesAndHdf(dir_path)
        createFolder(fullPath(dir_path, 'cp_results'))

        if len(imgList) == 0:
            if exists(fullPath(dir_path, 'cp_results/summary.csv')):
                self.browseHDF(dir_path, hdfList)
            else:
                errMsg = QMessageBox()
                errMsg.setText('No image and summary.csv detected')
                errMsg.setInformativeText('Please select an image or another folder to process.')
                errMsg.setStandardButtons(QMessageBox.Ok)
                errMsg.setIcon(QMessageBox.Warning)
                errMsg.exec_()
        else:
            df_sum = self.csvManager.df_sum
            df_rings = self.csvManager.df_rings
            imgs1 = set(df_sum['filename'])
            imgs2 = set(df_rings['filename'])
            # all_imgs = imgs1 & imgs2
            all_imgs = imgs1
            tmp_imlist = set(imgList[:])
            imgList = list(tmp_imlist - all_imgs)
            imgList.sort()
            if len(imgList) > 0:
                cp = CPImageWindow(self, "", dir_path, process_folder=True, imgList=imgList)
            self.browseHDF(dir_path, hdfList)

    def setMinMaxIntensity(self, img, minInt, maxInt, minIntLabel, maxIntLabel):
        min_val = img.min()
        max_val = img.max()
        self.intensityRange = [min_val, max_val - 1, min_val + 1, max_val]
        minInt.setMinimum(self.intensityRange[0])
        minInt.setMaximum(self.intensityRange[1])
        maxInt.setMinimum(self.intensityRange[2])
        maxInt.setMaximum(self.intensityRange[3])
        step = max(1., (max_val - min_val) / 100)
        minInt.setSingleStep(step)
        maxInt.setSingleStep(step)
        minIntLabel.setText("Min intensity (" + str(min_val) + ")")
        maxIntLabel.setText("Max intensity (" + str(max_val) + ")")

        if img.dtype == 'float32':
            decimal = 2
        else:
            decimal = 0

        maxInt.setDecimals(decimal)
        minInt.setDecimals(decimal)

        if maxInt.value() == 1. and minInt.value() == 0.:
            self.updatingUI = True
            minInt.setValue(min_val)
            maxInt.setValue(max_val * 0.6)
            self.updatingUI = False

    def updateRightStatusBar(self, text):
        QApplication.processEvents()
        self.statusLabel.setHidden(False)
        self.statusLabel.setText(text)
        QApplication.processEvents()

    def updateStatusBar(self, text, bar=None):
        QApplication.processEvents()
        self.imagePathLabel.setText(text)
        if bar is not None:
            self.progressBar.setValue(bar)
        QApplication.processEvents()
